import logging
from inflection import underscore
from collections import OrderedDict as _o
import gilda
from indra.sources import trips
from indra.ontology.standardize import \
    standardize_agent_name
from indra.statements.statements import *
from indra.assemblers.english.assembler import _assemble_agent_str, \
    EnglishAssembler, statement_base_verb, statement_present_verb
from indra.assemblers.pybel.assembler import _get_agent_node
from bioagents.tra.tra import MolecularQuantity, TemporalPattern
from .util import get_class_from_name


logger = logging.getLogger(__name__)


class Query(object):
    """The parent class of all query types."""
    @classmethod
    def _from_json(cls, json_dict):
        query_type = json_dict.get('type')
        query_cls = get_class_from_name(query_type, Query)
        query = query_cls._from_json(json_dict)
        return query

    def matches(self, other):
        return self.matches_key() == other.matches_key()

    def matches_key(self):
        pass

    def get_hash(self):
        return make_hash(self.matches_key(), 14)

    def get_hash_with_model(self, model_name):
        key = (self.matches_key(), model_name)
        return make_hash(mk_str(key), 14)

    def get_type(self):
        return underscore(type(self).__name__)


class StructuralProperty(Query):
    pass


class PathProperty(Query):
    """This type of query requires finding a mechanistic causally consistent
    path that satisfies query statement.

    Parameters
    ----------
    path_stmt : indra.statements.Statement
        A path to look for in the model represented as INDRA statement.
    entity_constraints : dict(list(indra.statements.Agent))
        A dictionary containing lists of Agents to be included in or excluded
        from the path.
    relationship_constraints : dict(list(str))
        A dictionary containing lists of Statement types to include in or
        exclude from the path.
    """
    def __init__(self, path_stmt, entity_constraints=None,
                 relationship_constraints=None):
        self.path_stmt = path_stmt
        if entity_constraints:
            self.include_entities = entity_constraints.get('include', [])
            self.exclude_entities = entity_constraints.get('exclude', [])
        else:
            self.include_entities = []
            self.exclude_entities = []
        if relationship_constraints:
            self.include_rels = relationship_constraints.get('include', [])
            self.exclude_rels = relationship_constraints.get('exclude', [])
        else:
            self.include_rels = []
            self.exclude_rels = []
        self.entities = self.get_entities()

    def to_json(self):
        query_type = self.get_type()
        json_dict = _o(type=query_type)
        json_dict['path'] = self.path_stmt.to_json()
        json_dict['entity_constraints'] = {}
        if self.include_entities:
            json_dict['entity_constraints']['include'] = [
                ec.to_json() for ec in self.include_entities]
        if self.exclude_entities:
            json_dict['entity_constraints']['exclude'] = [
                ec.to_json() for ec in self.exclude_entities]
        json_dict['relationship_constraints'] = {}
        if self.include_rels:
            json_dict['relationship_constraints']['include'] = [
                {'type': rel} for rel in self.include_rels]
        if self.exclude_rels:
            json_dict['relationship_constraints']['exclude'] = [
                {'type': rel} for rel in self.exclude_rels]
        return json_dict

    @classmethod
    def _from_json(cls, json_dict):
        path_stmt_json = json_dict.get('path')
        path_stmt = Statement._from_json(path_stmt_json)
        ent_constr_json = json_dict.get('entity_constraints')
        entity_constraints = None
        if ent_constr_json:
            entity_constraints = {}
            for key, value in ent_constr_json.items():
                entity_constraints[key] = [Agent._from_json(ec) for ec
                                           in value]
        rel_constr_json = json_dict.get('relationship_constraints')
        relationship_constraints = None
        if rel_constr_json:
            relationship_constraints = {}
            for key, value in rel_constr_json.items():
                relationship_constraints[key] = [
                    rel_type['type'] for rel_type in value]
        query = cls(path_stmt, entity_constraints, relationship_constraints)
        return query

    def get_entities(self):
        """Return entities from the path statement and the inclusion list."""
        path_entities = self.path_stmt.agent_list()
        return path_entities + self.include_entities

    def matches_key(self):
        key = self.path_stmt.matches_key()
        if self.include_entities:
            for ent in sorted(self.include_entities,
                              key=lambda x: x.matches_key()):
                key += ent.matches_key()
        if self.exclude_entities:
            for ent in sorted(self.exclude_entities,
                              key=lambda x: x.matches_key()):
                key += ent.matches_key()
        if self.include_rels:
            for rel in sorted(self.include_rels):
                key += rel
        if self.exclude_rels:
            for rel in sorted(self.exclude_rels):
                key += rel
        return mk_str(key)

    def __str__(self):
        parts = [f'PathPropertyQuery(stmt={str(self.path_stmt)}.']
        if self.include_entities:
            inents = ', '.join([str(e) for e in self.include_entities])
            parts.append(f' Include entities: {inents}.')
        if self.exclude_entities:
            exents = ', '.join([str(e) for e in self.exclude_entities])
            parts.append(f' Exclude entities: {exents}.')
        if self.include_rels:
            inrels = ', '.join(self.include_rels)
            parts.append(f' Include relations: {inrels}.')
        if self.exclude_rels:
            exrels = ', '.join(self.exclude_rels)
            parts.append(f' Exclude relations: {exrels}.')
        return ''.join(parts)

    def __repr__(self):
        return str(self)

    def to_english(self):
        ea = EnglishAssembler([self.path_stmt])
        return ea.make_model()


class SimpleInterventionProperty(Query):
    pass


class ComparativeInterventionProperty(Query):
    pass


class DynamicProperty(Query):
    """This type of query requires dynamic simulation of the model to check
    whether the queried temporal pattern is satisfied.

    Parameters
    ----------
    entity : indra.statements.Agent
        An entity to simulate the model for.
    pattern_type : str
        Type of temporal pattern. Accepted values: 'always_value', 'no_change',
        'eventual_value', 'sometime_value', 'sustained', 'transient'.
    quant_value : str or float
        Value of molecular quantity of entity of interest. Can be 'high' or
        'low' or a specific number.
    quant_type : str
        Type of molecular quantity of entity of interest. Default: qualitative.
    """
    def __init__(self, entity, pattern_type, quant_value=None,
                 quant_type='qualitative'):
        self.entity = entity
        self.pattern_type = pattern_type
        self.quant_value = quant_value
        self.quant_type = quant_type

    def get_temporal_pattern(self):
        """Return TemporalPattern object created with query properties."""
        mq = None
        if self.quant_value:
            mq = MolecularQuantity(self.quant_type, self.quant_value)
        tp = TemporalPattern(self.pattern_type, [self.entity], None, value=mq)
        return tp

    def matches_key(self):
        ent_matches_key = self.entity.matches_key()
        key = (ent_matches_key, self.pattern_type, self.quant_type,
               str(self.quant_value))
        return str(key)

    def to_json(self):
        query_type = self.get_type()
        json_dict = _o(type=query_type)
        json_dict['entity'] = self.entity.to_json()
        json_dict['pattern_type'] = self.pattern_type
        json_dict['quantity'] = {}
        json_dict['quantity']['type'] = self.quant_type
        json_dict['quantity']['value'] = self.quant_value
        return json_dict

    @classmethod
    def _from_json(cls, json_dict):
        ent_json = json_dict.get('entity')
        entity = Agent._from_json(ent_json)
        pattern_type = json_dict.get('pattern_type')
        quant_json = json_dict.get('quantity')
        quant_type = quant_json.get('type')
        quant_value = quant_json.get('value')
        query = cls(entity, pattern_type, quant_value, quant_type)
        return query

    def __str__(self):
        descr = (f'DynamicPropertyQuery(entity={self.entity}, '
                 f'pattern={self.pattern_type}, '
                 f'molecular quantity={(self.quant_type, self.quant_value)})')
        return descr

    def __repr__(self):
        return str(self)

    def to_english(self):
        agent = _assemble_agent_str(self.entity).agent_str
        agent = agent[0].upper() + agent[1:]
        if self.pattern_type == 'always_value':
            pattern = 'always'
        elif self.pattern_type == 'eventual_value':
            pattern = 'eventually'
        elif self.pattern_type == 'sometime_value':
            pattern = 'sometimes'
        elif self.pattern_type == 'no_change':
            pattern = 'not changing'
        else:
            pattern = self.pattern_type
        if self.quant_value:
            return f'{agent} is {pattern} {self.quant_value}.'
        return f'{agent} is {pattern}.'


class OpenSearchQuery(Query):
    """This type of query requires doing an open ended breadth-first search
    to find paths satisfying the query.

    Parameters
    ----------
    entity : indra.statements.Agent
        An entity to simulate the model for.
    stmt_type : str
        Name of statement type.
    entity_role : str
        What role entity should play in statement (subject or object).
    terminal_ns : list[str]
        Force a path to terminate when any of the namespaces in this list
        are encountered and only yield paths that terminate at these
        namepsaces

    Attributes
    ----------
    path_stmt : indra.statements.Statement
        An INDRA statement having its subject or object set to None to
        represent open search query.
    """
    def __init__(self, entity, stmt_type, entity_role, terminal_ns=None):
        self.entity = entity
        self.stmt_type = stmt_type
        self.entity_role = entity_role
        self.terminal_ns = terminal_ns
        self.path_stmt = self.make_stmt()

    def make_stmt(self):
        stmt_type = self.stmt_type
        if self.entity_role == 'subject':
            if self.stmt_type == 'IncreaseAmount':
                stmt_type = 'Activation'
            elif self.stmt_type == 'DecreaseAmount':
                stmt_type = 'Inhibition'
        stmt_class = get_statement_by_name(stmt_type)
        if self.entity_role == 'subject':
            subj = self.entity
            obj = None
        elif self.entity_role == 'object':
            subj = None
            obj = self.entity
        stmt = stmt_class(subj, obj)
        return stmt

    def get_sign(self, mc_type):
        if mc_type == 'unsigned_graph' or self.entity_role == 'object':
            sign = 0
        elif isinstance(self.path_stmt, RegulateActivity):
            sign = 0 if self.path_stmt.is_activation else 1
        elif isinstance(self.path_stmt, RegulateAmount):
            sign = 1 if isinstance(self.path_stmt, DecreaseAmount) else 0
        else:
            raise ValueError('Could not determine sign')
        return sign

    def matches_key(self):
        key = self.entity.matches_key()
        key += self.stmt_type
        key += self.entity_role
        if self.terminal_ns:
            for ns in self.terminal_ns:
                key += ns
        return mk_str(key)

    def to_json(self):
        query_type = self.get_type()
        json_dict = _o(type=query_type)
        json_dict['entity'] = self.entity.to_json()
        json_dict['stmt_type'] = self.stmt_type
        json_dict['entity_role'] = self.entity_role
        json_dict['terminal_ns'] = self.terminal_ns
        return json_dict

    @classmethod
    def _from_json(cls, json_dict):
        ent_json = json_dict.get('entity')
        entity = Agent._from_json(ent_json)
        stmt_type = json_dict.get('stmt_type')
        entity_role = json_dict.get('entity_role')
        terminal_ns = json_dict.get('terminal_ns')
        query = cls(entity, stmt_type, entity_role, terminal_ns)
        return query

    def __str__(self):
        parts = [f'OpenSearchQuery(stmt={self.path_stmt}.']
        if self.terminal_ns:
            parts.append(f' Terminal namespace={self.terminal_ns}')
        return ''.join(parts)

    def __repr__(self):
        return str(self)

    def to_english(self):
        agent = _assemble_agent_str(self.entity).agent_str
        if self.entity_role == 'subject':
            verb = statement_base_verb(self.stmt_type.lower())
            verb = verb[0].lower() + verb[1:]
            sentence = f'What does {agent} {verb}?'
        elif self.entity_role == 'object':
            verb = statement_present_verb(self.stmt_type.lower())
            verb = verb[0].lower() + verb[1:]
            sentence = f'What {verb} {agent}?'
        sentence = sentence[0].upper() + sentence[1:]
        if self.terminal_ns:
            sentence += f' ({", ".join(self.terminal_ns).upper()})'
        return sentence

    def get_entities(self):
        return [self.entity]


# This is the general method to get a grounding agent from text but it doesn't
# handle agent state which is required for dynamic queries
def get_agent_from_gilda(ag_name):
    """Return an INDRA Agent object by grounding its entity text with Gilda."""
    matches = gilda.ground(ag_name)
    if not matches:
        raise GroundingError(
            f"Could not find grounding for {ag_name} with Gilda.")
    agent = Agent(ag_name,
                  db_refs={'TEXT': ag_name,
                           matches[0].term.db: matches[0].term.id})
    standardize_agent_name(agent, standardize_refs=True)
    return agent


# This is the method that dynamical queries use to represent agents with
# state
def get_agent_from_trips(ag_text, service_host='http://34.230.33.149:8002/cgi/'):
    """Return an INDRA Agent object by grounding its entity text with TRIPS."""
    tp = trips.process_text(ag_text, service_host=service_host)
    agent_list = tp.get_agents()
    if not agent_list:
        raise GroundingError(
            f"Could not find grounding for {ag_text} with TRIPS.")
    return agent_list[0]


def get_agent_from_text(ag_text):
    """
    Return an INDRA Agent object by grounding its entity text with either
    Gilda or TRIPS.
    """
    try:
        agent = get_agent_from_gilda(ag_text)
        logger.info('Got agent from Gilda')
    except GroundingError:
        try:
            agent = get_agent_from_trips(ag_text)
            logger.info('Got agent from TRIPS')
        except GroundingError:
            raise GroundingError(f'Could not find grounding for {ag_text}.')
    return agent


class GroundingError(Exception):
    pass



import json
from os.path import abspath, dirname, join
from indra.statements import Phosphorylation, Agent, ModCondition, Inhibition
from emmaa.queries import Query, PathProperty, DynamicProperty, \
    OpenSearchQuery, get_agent_from_text, get_agent_from_trips, \
    get_agent_from_gilda


def test_path_property_from_json():
    query_file = join(dirname(abspath(__file__)), 'path_property_query.json')
    with open(query_file, 'r') as f:
        json_dict = json.load(f)
    query = Query._from_json(json_dict)
    assert query
    assert isinstance(query, PathProperty)
    assert isinstance(query.path_stmt, Phosphorylation), query.path_stmt
    assert query.path_stmt.enz.name == 'EGFR', query.path_stmt
    assert query.path_stmt.sub.name == 'ERK', query.path_stmt
    assert isinstance(query.exclude_entities[0], Agent)
    assert query.exclude_entities[0].name == 'PI3K'
    assert isinstance(query.include_entities[0], Agent)
    assert query.include_entities[0].name == 'MAPK1'
    assert set(query.exclude_rels) == set(['IncreaseAmount', 'DecreaseAmount'])
    assert query.include_rels[0] == 'Inhibition'


def test_path_property_to_json():
    stmt = Phosphorylation(enz=Agent('EGFR', db_refs={'HGNC': '3236'}),
                           sub=Agent('ERK', db_refs={'FPLX': 'ERK'}))
    entity_constraints = {'exclude': [Agent('PI3K', db_refs={'FPLX': 'PI3K'})]}
    relationship_contraints = {'exclude': ['IncreaseAmount', 'DecreaseAmount']}
    query = PathProperty(stmt, entity_constraints, relationship_contraints)
    assert query
    json = query.to_json()
    assert json.get('type') == 'path_property'
    path = json.get('path')
    assert path.get('type') == 'Phosphorylation'
    deserialize_query = Query._from_json(json)
    json2 = deserialize_query.to_json()
    assert json == json2, {'json': json, 'json2': json2}


def test_stringify_path_property():
    stmt = Phosphorylation(enz=Agent('EGFR', db_refs={'HGNC': '3236'}),
                           sub=Agent('ERK', db_refs={'FPLX': 'ERK'}))
    entity_constraints = {'exclude': [Agent('PI3K', db_refs={'FPLX': 'PI3K'})]}
    relationship_contraints = {'exclude': ['IncreaseAmount', 'DecreaseAmount']}
    query = PathProperty(stmt, entity_constraints, relationship_contraints)
    query_str = str(query)
    assert query_str == (
        'PathPropertyQuery(stmt=Phosphorylation(EGFR(), ERK()). Exclude '
        'entities: PI3K(). Exclude relations: IncreaseAmount, DecreaseAmount.')


def test_path_property_to_english():
    stmt = Phosphorylation(enz=Agent('EGFR', db_refs={'HGNC': '3236'}),
                           sub=Agent('ERK', db_refs={'FPLX': 'ERK'}))
    query = PathProperty(stmt)
    engl = query.to_english()
    assert engl == 'EGFR phosphorylates ERK.'


def test_dynamic_property_from_json():
    query_file = join(
        dirname(abspath(__file__)), 'dynamic_property_query.json')
    with open(query_file, 'r') as f:
        json_dict = json.load(f)
    query = Query._from_json(json_dict)
    assert query
    assert isinstance(query, DynamicProperty)
    assert isinstance(query.entity, Agent)
    assert query.entity.name == 'EGFR'
    assert isinstance(query.pattern_type, str)
    assert query.pattern_type == 'always_value'
    assert isinstance(query.quant_value, str)
    assert query.quant_value == 'low'
    assert isinstance(query.quant_type, str)
    assert query.quant_type == 'qualitative'


def test_dynamic_property_to_json():
    agent = Agent('EGFR', mods=[ModCondition('phosphorylation')],
                  db_refs={'HGNC': '3236'})
    query = DynamicProperty(agent, 'always_value', 'low', 'qualitative')
    json = query.to_json()
    assert json.get('type') == 'dynamic_property'
    entity = json.get('entity')
    assert entity.get('name') == 'EGFR'
    assert entity.get('db_refs') == {"HGNC": "3236"}
    assert json.get('pattern_type') == 'always_value'
    quantity = json.get('quantity')
    assert quantity.get('type') == 'qualitative'
    assert quantity.get('value') == 'low'


def test_stringify_dynamic_property():
    agent = Agent('EGFR', mods=[ModCondition('phosphorylation')],
                  db_refs={'HGNC': '3236'})
    query = DynamicProperty(agent, 'always_value', 'low', 'qualitative')
    query_str = str(query)
    assert query_str == ("DynamicPropertyQuery(entity=EGFR(mods: "
                         "(phosphorylation)), pattern=always_value, "
                         "molecular quantity=('qualitative', 'low'))")


def test_dynamic_property_to_english():
    agent = Agent('EGFR', mods=[ModCondition('phosphorylation')],
                  db_refs={'HGNC': '3236'})
    query = DynamicProperty(agent, 'always_value', 'low', 'qualitative')
    assert query.to_english() == 'Phosphorylated EGFR is always low.'
    query.pattern_type = 'eventual_value'
    assert query.to_english() == 'Phosphorylated EGFR is eventually low.'


def test_open_query_from_json():
    query_file = join(dirname(abspath(__file__)), 'open_query.json')
    with open(query_file, 'r') as f:
        json_dict = json.load(f)
    query = Query._from_json(json_dict)
    assert query
    assert isinstance(query, OpenSearchQuery)
    assert isinstance(query.entity, Agent)
    assert query.entity.name == 'EGFR'
    assert query.entity_role == 'object'
    assert query.stmt_type == 'Inhibition'
    assert isinstance(query.path_stmt, Inhibition)
    assert query.path_stmt.subj is None
    assert query.terminal_ns == ['chebi']


def test_open_query_to_json():
    ag = Agent('EGFR', db_refs={'HGNC': '3236'})
    query = OpenSearchQuery(ag, 'Inhibition', 'object', ['chebi'])
    assert query
    json = query.to_json()
    assert json.get('type') == 'open_search_query'
    assert json.get('stmt_type') == 'Inhibition'
    assert json.get('entity_role') == 'object'
    assert json.get('terminal_ns') == ['chebi']
    deserialize_query = Query._from_json(json)
    json2 = deserialize_query.to_json()
    assert json == json2, {'json': json, 'json2': json2}


def test_stringify_open_query():
    ag = Agent('EGFR', db_refs={'HGNC': '3236'})
    query = OpenSearchQuery(ag, 'Inhibition', 'object', ['chebi'])
    query_str = str(query)
    assert query_str == (
        "OpenSearchQuery(stmt=Inhibition(None, EGFR()). Terminal namespace="
        "['chebi']")


def test_open_query_to_english():
    ag = Agent('EGFR', db_refs={'HGNC': '3236'})
    q1 = OpenSearchQuery(ag, 'Inhibition', 'object', ['chebi', 'chembl'])
    q2 = OpenSearchQuery(ag, 'Inhibition', 'subject', ['hgnc'])
    q3 = OpenSearchQuery(ag, 'Activation', 'subject')
    assert q1.to_english() == 'What inhibits EGFR? (CHEBI, CHEMBL)', q1.to_english()
    assert q2.to_english() == 'What does EGFR inhibit? (HGNC)'
    assert q3.to_english() == 'What does EGFR activate?'


def test_get_sign():
    ag = Agent('EGFR', db_refs={'HGNC': '3236'})
    # When entity role is object, sign is always 0 (sign of upstream node)
    query = OpenSearchQuery(ag, 'Inhibition', 'object', ['chebi'])
    for mc_type in ['pysb', 'pybel', 'signed_graph', 'unsigned_graph']:
        assert query.get_sign(mc_type) == 0
    # Entity role is subject
    # Always 0 for unsigned graph
    query = OpenSearchQuery(ag, 'Inhibition', 'subject', ['chebi'])
    assert query.get_sign('unsigned_graph') == 0
    # For others, depends on statement type
    for mc_type in ['pysb', 'pybel', 'signed_graph']:
        assert query.get_sign(mc_type) == 1
    query = OpenSearchQuery(ag, 'Activation', 'subject', ['chebi'])
    for mc_type in ['pysb', 'pybel', 'signed_graph']:
        assert query.get_sign(mc_type) == 0


def test_grounding_from_gilda():
    agent = get_agent_from_gilda('MAPK1')
    assert isinstance(agent, Agent)
    assert agent.name == 'MAPK1'
    assert agent.db_refs == {'TEXT': 'MAPK1', 'HGNC': '6871', 'UP': 'P28482',
                             'MESH': 'D019950'}, agent.db_refs

    # test with lower case
    agent = get_agent_from_gilda('mapk1')
    assert isinstance(agent, Agent)
    assert agent.name == 'MAPK1'
    assert agent.db_refs == {'TEXT': 'mapk1', 'HGNC': '6871', 'UP': 'P28482',
                             'MESH': 'D019950'}, agent.db_refs
    # other agent
    agent = get_agent_from_gilda('BRAF')
    assert isinstance(agent, Agent)
    assert agent.name == 'BRAF'
    assert agent.db_refs == {'TEXT': 'BRAF',
                             'HGNC': '1097', 'UP': 'P15056'}, agent.db_refs


def test_agent_from_trips():
    ag = get_agent_from_trips('MAP2K1')
    assert isinstance(ag, Agent)
    assert ag.name == 'MAP2K1'
    assert not ag.mods
    ag_phos = get_agent_from_trips('the phosphorylated MAP2K1')
    assert ag_phos.mods


def test_generic_agent_from_text():
    agent = get_agent_from_gilda('MAPK1')
    assert isinstance(agent, Agent)
    assert agent.name == 'MAPK1'
    assert agent.db_refs == {'TEXT': 'MAPK1', 'HGNC': '6871', 'UP': 'P28482',
                             'MESH': 'D019950'}, agent.db_refs
    assert not agent.mods
    ag_phos = get_agent_from_trips('the phosphorylated MAP2K1')
    assert ag_phos.db_refs.get('HGNC') == '6840'
    assert ag_phos.mods



import datetime
from indra_db.client.principal.raw_statements import \
    get_raw_stmt_jsons_from_papers
from indra_db.util import get_db
from indra.statements import stmts_from_json
from emmaa.statements import EmmaaStatement


def read_db_ids_search_terms(id_search_terms, id_type):
    """Return extracted EmmaaStatements from INDRA database given an
    ID-search term dict.

    Parameters
    ----------
    id_search_terms : dict
        A dict representing a set of IDs pointing to search terms that
        produced them.

    Returns
    -------
    list[:py:class:`emmaa.model.EmmaaStatement`]
        A list of EmmaaStatements extracted from the given IDs.
    """
    ids = list(id_search_terms.keys())
    date = datetime.datetime.utcnow()
    db = get_db('primary')
    id_stmts = get_raw_stmt_jsons_from_papers(ids, id_type=id_type, db=db)
    estmts = []
    for _id, stmt_jsons in id_stmts.items():
        stmts = stmts_from_json(stmt_jsons)
        for stmt in stmts:
            es = EmmaaStatement(stmt, date, id_search_terms[_id])
            estmts.append(es)
    return estmts


def read_db_pmid_search_terms(pmid_search_terms):
    """Return extracted EmmaaStatements from INDRA database given a
    PMID-search term dict.

    Parameters
    ----------
    pmid_search_terms : dict
        A dict representing a set of PMIDs pointing to search terms that
        produced them.

    Returns
    -------
    list[:py:class:`emmaa.model.EmmaaStatement`]
        A list of EmmaaStatements extracted from the given PMIDs.
    """
    return read_db_ids_search_terms(pmid_search_terms, 'pmid')


def read_db_doi_search_terms(doi_search_terms):
    """Return extracted EmmaaStatements from INDRA database given a
    DOI-search term dict.

    Parameters
    ----------
    doi_search_terms : dict
        A dict representing a set of DOIs pointing to search terms that
        produced them.

    Returns
    -------
    list[:py:class:`emmaa.model.EmmaaStatement`]
        A list of EmmaaStatements extracted from the given DOIs.
    """
    return read_db_ids_search_terms(doi_search_terms, 'doi')



"""The AWS Lambda emmaa-model-stats definition.

This file contains the function that will be run when Lambda is triggered. It
must be placed on s3, which can either be done manually (not recommended) or
by running:

$ python update_lambda.py model_stats.py emmaa-model-stats

in this directory.
"""

import boto3
from datetime import datetime

JOB_DEF = 'emmaa_jobdef'
QUEUE = 'emmaa-models-update-test'
PROJECT = 'aske'
PURPOSE = 'update-emmaa-model-stats'
BRANCH = 'origin/master'


def lambda_handler(event, context):
    """Create a batch job to generate model statistics.

    This function is designed to be placed on lambda, taking the event and
    context arguments that are passed. Event parameter is used here to pass
    name of the model.

    This Lambda function is configured to be invoked by emmaa-after-update
    Lambda function.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event. In
        this case the dictionary contains 'model' key.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A dict containing 'statusCode', with a valid HTTP status code, and any
        other data to be returned to Lambda.
    """
    batch = boto3.client('batch')
    model_name = event['model']
    core_command = 'bash scripts/git_and_run.sh'
    if BRANCH is not None:
        core_command += f' --branch {BRANCH}'
    core_command += (' python scripts/run_model_stats_from_s3.py'
                        f' --model {model_name}  --stats_mode model')
    print(core_command)
    cont_overrides = {
        'command': ['python', '-m', 'indra.util.aws', 'run_in_batch',
                    '--project', PROJECT, '--purpose', PURPOSE,
                    core_command]
        }
    now_str = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
    ret = batch.submit_job(
        jobName=f'{model_name}_model_stats_{now_str}',
        jobQueue=QUEUE, jobDefinition=JOB_DEF,
        containerOverrides=cont_overrides)
    job_id = ret['jobId']

    return {'statusCode': 200, 'result': 'SUCCESS', 'job_id': job_id}



"""This module implements the object model for EMMAA model testing."""
import logging
import itertools
import jsonpickle
import os
import pickle
import sys
from collections import defaultdict
from fnvhash import fnv1a_32
from urllib import parse
from copy import deepcopy
from zipfile import ZipFile
from indra.explanation.model_checker import PysbModelChecker, \
    PybelModelChecker, SignedGraphModelChecker, UnsignedGraphModelChecker
from indra.explanation.reporting import stmts_from_pysb_path, \
    stmts_from_pybel_path, stmts_from_indranet_path, PybelEdge, \
    pybel_edge_to_english
from indra.explanation.pathfinding import bfs_search_multiple_nodes
from indra.assemblers.english.assembler import EnglishAssembler
from indra.sources.indra_db_rest.api import get_statement_queries
from indra.statements import Statement, Agent, Concept, Event, \
    stmts_to_json
from indra.util.statement_presentation import group_and_sort_statements
from bioagents.tra.tra import TRA, MissingMonomerError, MissingMonomerSiteError
from emmaa.model import EmmaaModel, get_assembled_statements
from emmaa.queries import PathProperty, DynamicProperty, OpenSearchQuery
from emmaa.util import make_date_str, get_s3_client, get_class_from_name, \
    EMMAA_BUCKET_NAME, find_latest_s3_file, load_pickle_from_s3, \
    save_pickle_to_s3, load_json_from_s3, save_json_to_s3, strip_out_date, \
    save_zip_json_to_s3


logger = logging.getLogger(__name__)
sys.setrecursionlimit(50000)


result_codes_link = ('https://emmaa.readthedocs.io/en/latest/dashboard/'
                     'response_codes.html')
RESULT_CODES = {
    'STATEMENT_TYPE_NOT_HANDLED': 'Statement type not handled',
    'SUBJECT_MONOMERS_NOT_FOUND': 'Statement subject not in model',
    'SUBJECT_NOT_FOUND': 'Statement subject not in model',
    'OBSERVABLES_NOT_FOUND': 'Statement object state not in model',
    'OBJECT_NOT_FOUND': 'Statement object state not in model',
    'NO_PATHS_FOUND': 'No path found that satisfies the test statement',
    'MAX_PATH_LENGTH_EXCEEDED': 'Path found but exceeds search depth',
    'PATHS_FOUND': 'Path found which satisfies the test statement',
    'INPUT_RULES_NOT_FOUND': 'No rules with test statement subject',
    'MAX_PATHS_ZERO': 'Path found but not reconstructed',
    'QUERY_NOT_APPLICABLE': 'Query is not applicable for this model',
    'NODE_NOT_FOUND': 'Node not in model'
}
ARROW_DICT = {'Complex': u"\u2194",
              'Inhibition': u"\u22A3",
              'DecreaseAmount': u"\u22A3"}


class ModelManager(object):
    """Manager to generate and store properties of a model and relevant tests.

    Parameters
    ----------
    model : emmaa.model.EmmaaModel
        EMMAA model

    Attributes
    ----------
    mc_mapping : dict
        A dictionary mapping a ModelChecker type to a corresponding method
        for assembling the model and a ModelChecker class.
    mc_types : dict
        A dictionary in which each key is a type of a ModelChecker and value is
        a dictionary containing an instance of a model, an instance of a
        ModelChecker and a list of test results.
    entities : list[indra.statements.agent.Agent]
        A list of entities of EMMAA model.
    applicable_tests : list[emmaa.model_tests.EmmaaTest]
        A list of EMMAA tests applicable for given EMMAA model.
    date_str : str
        Time when this object was created.
    """
    def __init__(self, model):
        self.model = model
        self.mc_mapping = {
            'pysb': (self.model.assemble_pysb, PysbModelChecker,
                     stmts_from_pysb_path),
            'pybel': (self.model.assemble_pybel, PybelModelChecker,
                      stmts_from_pybel_path),
            'signed_graph': (self.model.assemble_signed_graph,
                             SignedGraphModelChecker,
                             stmts_from_indranet_path),
            'unsigned_graph': (self.model.assemble_unsigned_graph,
                               UnsignedGraphModelChecker,
                               stmts_from_indranet_path)}
        self.mc_types = {}
        for mc_type in model.test_config.get('mc_types', ['pysb']):
            self.mc_types[mc_type] = {}
            assembled_model = self.mc_mapping[mc_type][0]()
            self.mc_types[mc_type]['model'] = assembled_model
            self.mc_types[mc_type]['model_checker'] = (
                self.mc_mapping[mc_type][1](assembled_model))
            self.mc_types[mc_type]['test_results'] = []
        self.entities = self.model.get_assembled_entities()
        self.applicable_tests = []
        self.date_str = make_date_str()
        self.path_stmt_counts = defaultdict(int)

    def get_updated_mc(self, mc_type, stmts, add_ns=False):
        """Update the ModelChecker and graph with stmts for tests/queries."""
        mc = self.mc_types[mc_type]['model_checker']
        mc.statements = stmts
        if mc_type == 'pysb':
            mc.graph = None
            mc.model_stmts = self.model.assembled_stmts
            mc.get_graph(prune_im=True, prune_im_degrade=True,
                         add_namespaces=add_ns)
        return mc

    def add_test(self, test):
        """Add a test to a list of applicable tests."""
        self.applicable_tests.append(test)

    def add_result(self, mc_type, result):
        """Add a result to a list of results."""
        self.mc_types[mc_type]['test_results'].append(result)

    def run_all_tests(self):
        """Run all applicable tests with all available ModelCheckers."""
        max_path_length, max_paths = self._get_test_configs()
        for mc_type in self.mc_types:
            self.run_tests_per_mc(mc_type, max_path_length, max_paths)

    def run_tests_per_mc(self, mc_type, max_path_length, max_paths):
        """Run all applicable tests with one ModelChecker."""
        mc = self.get_updated_mc(
            mc_type, [test.stmt for test in self.applicable_tests])
        logger.info(f'Running the tests with {mc_type} ModelChecker.')
        results = mc.check_model(
            max_path_length=max_path_length, max_paths=max_paths)
        for (stmt, result) in results:
            self.add_result(mc_type, result)

    def make_path_json(self, mc_type, result_paths):
        paths = []
        json_lines = []
        for path in result_paths:
            path_nodes = []
            edge_list = []
            path_node_list = []
            hashes = []
            report_function = self.mc_mapping[mc_type][2]
            model = self.mc_types[mc_type]['model']
            stmts = self.model.assembled_stmts
            if mc_type == 'pysb':
                report_stmts = report_function(path, model, stmts)
                path_stmts = [[st] for st in report_stmts]
                merge = False
            elif mc_type == 'pybel':
                path_stmts = report_function(path, model, False, stmts)
                merge = False
            elif mc_type == 'signed_graph':
                path_stmts = report_function(path, model, True, False, stmts)
                merge = True
            elif mc_type == 'unsigned_graph':
                path_stmts = report_function(path, model, False, False, stmts)
                merge = True
            for i, step in enumerate(path_stmts):
                edge_nodes = []
                if len(step) < 1:
                    continue
                stmt_type = type(step[0]).__name__
                if stmt_type == 'PybelEdge':
                    source, target = step[0].source, step[0].target
                    edge_nodes.append(source.name)
                    edge_nodes.append(u"\u2192")
                    edge_nodes.append(target.name)
                    hashes.append(['not a statement'])
                else:
                    step_hashes = []
                    for stmt in step:
                        self.path_stmt_counts[stmt.get_hash()] += 1
                        step_hashes.append(stmt.get_hash())
                    hashes.append(step_hashes)
                    agents = [ag.name if ag is not None else None
                              for ag in step[0].agent_list()]
                    # For complexes make sure that the agent from the
                    # previous edge goes first
                    if stmt_type == 'Complex' and len(path_nodes) > 0:
                        agents = sorted(
                            [ag for ag in agents if ag is not None],
                            key=lambda x: x != path_nodes[-1])
                    for j, ag in enumerate(agents):
                        if ag is not None:
                            edge_nodes.append(ag)
                        if j == (len(agents) - 1):
                            break
                        if stmt_type in ARROW_DICT:
                            edge_nodes.append(ARROW_DICT[stmt_type])
                        else:
                            edge_nodes.append(u"\u2192")
                if i == 0:
                    for n in edge_nodes:
                        path_nodes.append(n)
                    path_node_list.append(edge_nodes[0])
                    path_node_list.append(edge_nodes[-1])
                else:
                    for n in edge_nodes[1:]:
                        path_nodes.append(n)
                    path_node_list.append(edge_nodes[-1])
                step_sentences = self._make_path_stmts(step, merge=merge)
                edge_dict = {'edge': ' '.join(edge_nodes),
                             'stmts': step_sentences}
                edge_list.append(edge_dict)
            path_json = {'path': ' '.join(path_nodes),
                         'edge_list': edge_list}
            one_line_path_json = {'nodes': path_node_list, 'edges': hashes,
                                  'graph_type': mc_type}
            paths.append(path_json)
            json_lines.append(one_line_path_json)
        return paths, json_lines

    def _make_path_stmts(self, stmts, merge=False):
        sentences = []
        date = strip_out_date(self.date_str, 'date')
        if merge:
            groups = group_and_sort_statements(stmts)
            for group in groups:
                group_stmts = group[-1]
                stmt_type = group[0][-1]
                agent_names = group[0][1]
                if len(agent_names) < 2:
                    continue
                if stmt_type == 'Influence':
                    stmt = get_class_from_name(stmt_type, Statement)(
                        Event(Concept(agent_names[0])),
                        Event(Concept(agent_names[1])))
                elif stmt_type == 'Conversion':
                    stmt = get_class_from_name(stmt_type, Statement)(
                        Agent(agent_names[0]),
                        [Agent(ag) for ag in agent_names[1]],
                        [Agent(ag) for ag in agent_names[2]])
                else:
                    try:
                        stmt = get_class_from_name(stmt_type, Statement)(
                            Agent(agent_names[0]), Agent(agent_names[1]))
                    except ValueError:
                        stmt = get_class_from_name(stmt_type, Statement)(
                            [Agent(ag_name) for ag_name in agent_names])
                ea = EnglishAssembler([stmt])
                sentence = ea.make_model()
                stmt_hashes = [gr_st.get_hash() for gr_st in group_stmts]
                url_param = parse.urlencode(
                    {'stmt_hash': stmt_hashes, 'source': 'model_statement',
                     'model': self.model.name, 'date': date}, doseq=True)
                link = f'/evidence?{url_param}'
                sentences.append((link, sentence, ''))
        else:
            for stmt in stmts:
                if isinstance(stmt, PybelEdge):
                    sentence = pybel_edge_to_english(stmt)
                    sentences.append(('', sentence, ''))
                else:
                    ea = EnglishAssembler([stmt])
                    sentence = ea.make_model()
                    stmt_hashes = [stmt.get_hash()]
                    url_param = parse.urlencode(
                        {'stmt_hash': stmt_hashes, 'source': 'model_statement',
                         'model': self.model.name, 'date': date}, doseq=True)
                    link = f'/evidence?{url_param}'
                    sentences.append((link, sentence, ''))
        return sentences

    def make_result_code(self, result):
        result_code = result.result_code
        return RESULT_CODES[result_code]

    def answer_query(self, query, **kwargs):
        if isinstance(query, DynamicProperty):
            return self.answer_dynamic_query(query, **kwargs)
        if isinstance(query, PathProperty):
            return self.answer_path_query(query)
        if isinstance(query, OpenSearchQuery):
            return self.answer_open_query(query)

    def answer_path_query(self, query):
        """Answer user query with a path if it is found."""
        if ScopeTestConnector.applicable(self, query):
            results = []
            for mc_type in self.mc_types:
                mc = self.get_updated_mc(mc_type, [query.path_stmt])
                max_path_length, max_paths = self._get_test_configs(
                    mode='query', mc_type=mc_type, default_paths=5)
                result = mc.check_statement(
                    query.path_stmt, max_paths, max_path_length)
                results.append(
                    (mc_type, self.process_response(mc_type, result)))
            return results
        else:
            return [('', self.hash_response_list(
                RESULT_CODES['QUERY_NOT_APPLICABLE']))]

    def answer_dynamic_query(self, query, use_kappa=False,
                             bucket=EMMAA_BUCKET_NAME):
        """Answer user query by simulating a PySB model."""
        tra = TRA(use_kappa=use_kappa)
        tp = query.get_temporal_pattern()
        pysb_model = deepcopy(self.mc_types['pysb']['model'])
        try:
            sat_rate, num_sim, kpat, pat_obj, fig_path = tra.check_property(
                pysb_model, tp)
            fig_name, ext = os.path.splitext(os.path.basename(fig_path))
            date_str = make_date_str()
            s3_key = (f'query_images/{self.model.name}/{fig_name}_'
                      f'{date_str}{ext}')
            s3_path = f'https://{bucket}.s3.amazonaws.com/{s3_key}'
            client = get_s3_client(unsigned=False)
            logger.info(f'Uploading image to {s3_path}')
            client.upload_file(fig_path, Bucket=bucket, Key=s3_key)
            resp_json = {'sat_rate': sat_rate, 'num_sim': num_sim,
                         'kpat': kpat, 'fig_path': s3_path}
        except (MissingMonomerError, MissingMonomerSiteError):
            resp_json = RESULT_CODES['QUERY_NOT_APPLICABLE']
        return [('pysb', self.hash_response_list(resp_json))]

    def answer_open_query(self, query):
        """Answer user open search query with found paths."""
        if ScopeTestConnector.applicable(self, query):
            results = []
            for mc_type in self.mc_types:
                max_path_length, max_paths = self._get_test_configs(
                    mode='query', qtype='open_search', mc_type=mc_type,
                    default_paths=50, default_length=2)
                add_ns = False
                if query.terminal_ns:
                    add_ns = True
                mc = self.get_updated_mc(mc_type, [query.path_stmt], add_ns)
                res = self.open_query_per_mc(
                    mc_type, mc, query, max_path_length, max_paths)
                results.append((mc_type, res))
            return results
        else:
            return [('', self.hash_response_list(
                RESULT_CODES['QUERY_NOT_APPLICABLE']))]

    def open_query_per_mc(self, mc_type, mc, query, max_path_length,
                          max_paths):
        g = mc.get_graph()
        subj_nodes, obj_nodes, res_code = mc.get_all_subjects_objects(
            query.path_stmt)
        if res_code:
            return self.hash_response_list(RESULT_CODES[res_code])
        else:
            if query.entity_role == 'subject':
                reverse = False
                assert subj_nodes is not None
                nodes = subj_nodes
            else:
                reverse = True
                assert obj_nodes is not None
                nodes = obj_nodes
            sign = query.get_sign(mc_type)
            if mc_type == 'pysb':
                terminal_ns = None
            else:
                terminal_ns = query.terminal_ns
            paths_gen = bfs_search_multiple_nodes(
                g, nodes, reverse=reverse, terminal_ns=terminal_ns,
                depth_limit=max_path_length, path_limit=max_paths, sign=sign)
            paths = []
            for p in paths_gen:
                if reverse:
                    paths.append(p[::-1])
                else:
                    paths.append(p)
            return self.process_open_query_response(mc_type, paths)

    def answer_queries(self, queries, **kwargs):
        """Answer all queries registered for this model.

        Parameters
        ----------
        queries : list[emmaa.queries.Query]
            A list of queries to run.

        Returns
        -------
        responses : list[tuple(json, json)]
            A list of tuples each containing a query, mc_type and result json.
        """
        responses = []
        applicable_queries = []
        applicable_stmts = []
        applicable_open_queries = []
        applicable_open_stmts = []
        for query in queries:
            # Dynamic queries need to be answered individually, while for
            # path and open queries some parts can be shared
            if isinstance(query, DynamicProperty):
                mc_type, response = self.answer_dynamic_query(
                    query, **kwargs)[0]
                responses.append((query, mc_type, response))
            elif isinstance(query, PathProperty):
                if ScopeTestConnector.applicable(self, query):
                    applicable_queries.append(query)
                    applicable_stmts.append(query.path_stmt)
                else:
                    responses.append(
                        (query, '', self.hash_response_list(
                            RESULT_CODES['QUERY_NOT_APPLICABLE'])))
            elif isinstance(query, OpenSearchQuery):
                if ScopeTestConnector.applicable(self, query):
                    applicable_open_queries.append(query)
                    applicable_open_stmts.append(query.path_stmt)
                else:
                    responses.append(
                        (query, '', self.hash_response_list(
                            RESULT_CODES['QUERY_NOT_APPLICABLE'])))

        # Only do the following steps if there are applicable queries
        # Path queries
        if applicable_queries:
            for mc_type in self.mc_types:
                mc = self.get_updated_mc(mc_type, applicable_stmts)
                max_path_length, max_paths = self._get_test_configs(
                    mode='query', mc_type=mc_type, default_paths=5)
                results = mc.check_model(
                    max_path_length=max_path_length, max_paths=max_paths)
                for ix, (_, result) in enumerate(results):
                    responses.append(
                        (applicable_queries[ix], mc_type,
                         self.process_response(mc_type, result)))

        # Open queries
        if applicable_open_queries:
            for mc_type in self.mc_types:
                max_path_length, max_paths = self._get_test_configs(
                    mode='query', mc_type=mc_type, default_paths=50,
                    default_length=2)
                mc = self.get_updated_mc(mc_type, applicable_open_stmts, True)
                for query in applicable_open_queries:
                    res = self.open_query_per_mc(
                        mc_type, mc, query, max_path_length, max_paths)
                    responses.append((query, mc_type, res))

        return sorted(responses, key=lambda x: x[0].matches_key())

    def _get_test_configs(self, mode='test', qtype='statement_checking',
                          mc_type=None, default_length=5, default_paths=1):
        if mode == 'test':
            config = self.model.test_config
        elif mode == 'query':
            config = self.model.query_config
        try:
            max_path_length = \
                config[qtype][mc_type]['max_path_length']
        except KeyError:
            try:
                max_path_length = \
                    config[qtype]['max_path_length']
            except KeyError:
                max_path_length = default_length
        try:
            max_paths = \
                config[qtype][mc_type]['max_paths']
        except KeyError:
            try:
                max_paths = \
                    config[qtype]['max_paths']
            except KeyError:
                max_paths = default_paths
        logger.info('Parameters for model checking: %d, %d' %
                    (max_path_length, max_paths))
        return (max_path_length, max_paths)

    def process_response(self, mc_type, result):
        """Return a dictionary in which every key is a hash and value is a list
        of tuples. Each tuple contains a sentence describing either a step in a
        path (if it was found) or result code (if a path was not found) and a
        link leading to a webpage with more information about corresponding
        sentence.
        """
        if result.paths:
            response, _ = self.make_path_json(mc_type, result.paths)
        else:
            response = self.make_result_code(result)
        return self.hash_response_list(response)

    def process_open_query_response(self, mc_type, paths):
        if paths:
            response, _ = self.make_path_json(mc_type, paths)
        else:
            response = 'No paths found that satisfy this query'
        return self.hash_response_list(response)

    def hash_response_list(self, response):
        """Return a dictionary mapping a hash with a response in a response
        list.
        """
        response_dict = {}
        if isinstance(response, str):
            response_hash = str(fnv1a_32(response.encode('utf-8')))
            response_dict[response_hash] = response
        elif isinstance(response, list):
            for resp in response:
                sentences = []
                for edge in resp['edge_list']:
                    for (_, sentence, _) in edge['stmts']:
                        sentences.append(sentence)
                response_str = ' '.join(sentences)
                response_hash = str(fnv1a_32(response_str.encode('utf-8')))
                response_dict[response_hash] = resp
        elif isinstance(response, dict):
            results = [str(response.get('sat_rate')),
                       str(response.get('num_sim'))]
            response_str = ' '.join(results)
            response_hash = str(fnv1a_32(response_str.encode('utf-8')))
            response_dict[response_hash] = response
        else:
            raise TypeError('Response should be a string or a list.')
        return response_dict

    def results_to_json(self, test_data=None):
        """Put test results to json format."""
        pickler = jsonpickle.pickler.Pickler()
        results_json = []
        results_json.append({
            'model_name': self.model.name,
            'mc_types': [mc_type for mc_type in self.mc_types.keys()],
            'path_stmt_counts': self.path_stmt_counts,
            'date_str': self.date_str,
            'test_data': test_data})
        json_lines = []
        for ix, test in enumerate(self.applicable_tests):
            test_ix_results = {'test_type': test.__class__.__name__,
                               'test_json': test.to_json()}
            for mc_type in self.mc_types:
                result = self.mc_types[mc_type]['test_results'][ix]
                path_json, test_json_lines = self.make_path_json(
                    mc_type, result.paths)
                test_ix_results[mc_type] = {
                    'result_json': pickler.flatten(result),
                    'path_json': path_json,
                    'result_code': self.make_result_code(result)}
                for line in test_json_lines:
                    # Only include lines with paths
                    if line:
                        line.update({'test': test.stmt.get_hash()})
                        json_lines.append(line)
            results_json.append(test_ix_results)
        return results_json, json_lines

    def upload_results(self, test_corpus='large_corpus_tests',
                       test_data=None, bucket=EMMAA_BUCKET_NAME):
        """Upload results to s3 bucket."""
        json_dict, json_lines = self.results_to_json(test_data)
        result_key = (f'results/{self.model.name}/results_'
                      f'{test_corpus}_{self.date_str}.json')
        paths_key = (f'paths/{self.model.name}/paths_{test_corpus}_'
                     f'{self.date_str}.jsonl')
        latest_paths_key = (f'paths/{self.model.name}/{test_corpus}'
                            '_latest_paths.jsonl')
        logger.info(f'Uploading test results to {result_key}')
        save_json_to_s3(json_dict, bucket, result_key)
        logger.info(f'Uploading test paths to {paths_key}')
        save_json_to_s3(json_lines, bucket, paths_key, save_format='jsonl')
        save_json_to_s3(json_lines, bucket, latest_paths_key, 'jsonl')

    def save_assembled_statements(self, bucket=EMMAA_BUCKET_NAME):
        """Upload assembled statements jsons to S3 bucket."""
        stmts = self.model.assembled_stmts
        stmts_json = stmts_to_json(stmts)
        # Save a timestapmed version and a generic latest version of files
        dated_key = f'assembled/{self.model.name}/statements_{self.date_str}'
        latest_key = f'assembled/{self.model.name}/' \
                     f'latest_statements_{self.model.name}'
        for ext in ('json', 'jsonl'):
            latest_obj_key = latest_key + '.' + ext
            logger.info(f'Uploading assembled statements to {latest_obj_key}')
            save_json_to_s3(stmts_json, bucket, latest_obj_key, ext)
        dated_jsonl = dated_key + '.jsonl'
        dated_zip = dated_key + '.zip'
        logger.info(f'Uploading assembled statements to {dated_jsonl}')
        save_json_to_s3(stmts_json, bucket, dated_jsonl, 'jsonl')
        logger.info(f'Uploading assembled statements to {dated_zip}')
        save_zip_json_to_s3(stmts_json, bucket, dated_zip, 'json')


class TestManager(object):
    """Manager to generate and run a set of tests on a set of models.

    Parameters
    ----------
    model_managers : list[emmaa.model_tests.ModelManager]
        A list of ModelManager objects
    tests : list[emmaa.model_tests.EmmaaTest]
        A list of EMMAA tests
    """
    def __init__(self, model_managers, tests):
        self.model_managers = model_managers
        self.tests = tests

    def make_tests(self, test_connector):
        """Generate a list of applicable tests for each model with a given test
        connector.

        Parameters
        ----------
        test_connector : emmaa.model_tests.TestConnector
            A TestConnector object to use for connecting models to tests.
        """
        logger.info(f'Checking applicability of {len(self.tests)} tests to '
                    f'{len(self.model_managers)} models')
        for model_manager, test in itertools.product(self.model_managers,
                                                     self.tests):
            if test_connector.applicable(model_manager, test):
                model_manager.add_test(test)
                logger.debug(f'Test {test.stmt} is applicable')
            else:
                logger.debug(f'Test {test.stmt} is not applicable')
        logger.info(f'Created tests for {len(self.model_managers)} models.')
        for model_manager in self.model_managers:
            logger.info(f'Created {len(model_manager.applicable_tests)} tests '
                        f'for {model_manager.model.name} model.')

    def run_tests(self):
        """Run tests for a list of model-test pairs"""
        for model_manager in self.model_managers:
            model_manager.run_all_tests()


class TestConnector(object):
    """Determines if a given test is applicable to a given model."""
    def __init__(self):
        pass

    @staticmethod
    def applicable(model, test):
        """Return True if the test is applicable to the given model."""
        return True


class ScopeTestConnector(TestConnector):
    """Determines applicability of a test to a model by overlap in scope."""
    @staticmethod
    def applicable(model, test):
        """Return True of all test entities are in the set of model entities"""
        model_entities = model.entities
        test_entities = test.get_entities()
        # TODO
        # After adding entities as an attribute to StatementCheckingTest(), use
        # test_entities = test.entities
        return ScopeTestConnector._overlap(model_entities, test_entities)

    @staticmethod
    def _overlap(model_entities, test_entities):
        me_names = {e.name for e in model_entities}
        te_names = {e.name for e in test_entities}
        # If all test entities are in model entities, we get an empty set here
        # so we return True
        return not te_names - me_names


class EmmaaTest(object):
    """Represent an EMMAA test condition"""
    def get_entities(self):
        """Return a list of entities that the test checks for."""
        raise NotImplementedError()


class StatementCheckingTest(EmmaaTest):
    """Represent an EMMAA test condition that checks a PySB-assembled model
    against an INDRA Statement."""
    def __init__(self, stmt, configs=None):
        self.stmt = stmt
        self.configs = {} if not configs else configs
        logger.info('Test configs: %s' % configs)
        # TODO
        # Add entities as a property if we can reload tests on s3.
        # self.entities = self.get_entities()

    def check(self, model_checker, pysb_model):
        """Use a model checker to check if a given model satisfies the test."""
        max_path_length = self.configs.get('max_path_length', 5)
        max_paths = self.configs.get('max_paths', 1)
        logger.info('Parameters for model checking: %s, %d' %
                    (max_path_length, max_paths))
        res = model_checker.check_statement(
            self.stmt,
            max_path_length=max_path_length,
            max_paths=max_paths)
        return res

    def get_entities(self):
        """Return a list of entities that the test checks for."""
        return self.stmt.agent_list()

    def to_json(self):
        return self.stmt.to_json()

    def __repr__(self):
        return "%s(stmt=%s)" % (self.__class__.__name__, repr(self.stmt))


def load_tests_from_s3(test_name, bucket=EMMAA_BUCKET_NAME):
    """Load Emmaa Tests with the given name from S3.

    Parameters
    ----------
    test_name : str
        Looks for a test file in the emmaa bucket on S3 with key
        'tests/{test_name}'.

    Return
    ------
    list of EmmaaTest
        List of EmmaaTest objects loaded from S3.
    """
    prefix = f'tests/{test_name}'
    try:
        test_key = find_latest_s3_file(bucket, prefix, '.pkl')
    except ValueError:
        test_key = f'tests/{test_name}.pkl'
    logger.info(f'Loading tests from {test_key}')
    tests = load_pickle_from_s3(bucket, test_key)
    return tests, test_key


def save_model_manager_to_s3(model_name, model_manager,
                             bucket=EMMAA_BUCKET_NAME):
    logger.info(f'Saving a model manager for {model_name} model to S3.')
    date_str = model_manager.date_str
    model_manager.model.stmts = []
    model_manager.model.assembled_stmts = []
    save_pickle_to_s3(model_manager, bucket,
                      f'results/{model_name}/model_manager_{date_str}.pkl')


def load_model_manager_from_s3(model_name=None, key=None,
                               bucket=EMMAA_BUCKET_NAME):
    # First try find the file from specified key
    if key:
        try:
            model_manager = load_pickle_from_s3(bucket, key)
            if not model_manager.model.assembled_stmts:
                stmts, _ = get_assembled_statements(
                    model_manager.model.name,
                    strip_out_date(model_manager.date_str, 'date'),
                    bucket=bucket)
                model_manager.model.assembled_stmts = stmts
            return model_manager
        except Exception as e:
            logger.info('Could not load the model manager')
            logger.info(e)
    # Now try find the latest key for given model
    if model_name:
        # Versioned
        key = find_latest_s3_file(
            bucket, f'results/{model_name}/model_manager_', '.pkl')
        if key is None:
            # Non-versioned
            key = f'results/{model_name}/latest_model_manager.pkl'
        return load_model_manager_from_s3(key=key, bucket=bucket)
    # Could not find either from key or from model name.
    logger.info('Could not find the model manager.')
    return None


def update_model_manager_on_s3(model_name, bucket=EMMAA_BUCKET_NAME):
    model = EmmaaModel.load_from_s3(model_name, bucket=bucket)
    mm = ModelManager(model)
    save_model_manager_to_s3(model_name, mm, bucket=bucket)
    return mm


def model_to_tests(model_name, upload=True, bucket=EMMAA_BUCKET_NAME):
    em = EmmaaModel.load_from_s3(model_name, bucket=bucket)
    em.run_assembly()
    tests = [StatementCheckingTest(stmt) for stmt in em.assembled_stmts if
             all(stmt.agent_list())]
    date_str = make_date_str()
    test_description = (
        f'These tests were generated from the {em.human_readable_name} '
        f'on {date_str[:10]}')
    test_dict = {'test_data': {'description': test_description},
                 'tests': tests}
    if upload:
        save_tests_to_s3(test_dict, bucket,
                         f'tests/{model_name}_tests_{date_str}.pkl', 'pkl')
    return test_dict


def save_tests_to_s3(tests, bucket, key, save_format='pkl'):
    """Save tests in pkl, json or jsonl format."""
    if save_format == 'pkl':
        save_pickle_to_s3(tests, bucket, key)
    elif save_format in ['json', 'jsonl']:
        if isinstance(tests, list):
            stmts = [test.stmt for test in tests]
        elif isinstance(tests, dict):
            stmts = [test.stmt for test in tests['tests']]
        stmts_json = stmts_to_json(stmts)
        save_json_to_s3(stmts_json, bucket, key, save_format)


def run_model_tests_from_s3(model_name, test_corpus='large_corpus_tests',
                            upload_results=True, bucket=EMMAA_BUCKET_NAME):
    """Run a given set of tests on a given model, both loaded from S3.

    After loading both the model and the set of tests, model/test overlap
    is determined using a ScopeTestConnector and tests are run.


    Parameters
    ----------
    model_name : str
        Name of EmmaaModel to load from S3.
    test_corpus : str
        Name of the file containing tests on S3.
    upload_results : Optional[bool]
        Whether to upload test results to S3 in JSON format. Can be set
        to False when running tests. Default: True

    Returns
    -------
    emmaa.model_tests.ModelManager
        Instance of ModelManager containing the model data, list of applied
        tests and the test results.
    """
    mm = load_model_manager_from_s3(model_name=model_name, bucket=bucket)
    test_dict, _ = load_tests_from_s3(test_corpus, bucket=bucket)
    if isinstance(test_dict, dict):
        tests = test_dict['tests']
        test_data = test_dict['test_data']
    elif isinstance(test_dict, list):
        tests = test_dict
        test_data = None
    tm = TestManager([mm], tests)
    tm.make_tests(ScopeTestConnector())
    tm.run_tests()
    # Optionally upload test results to S3
    if upload_results:
        mm.upload_results(test_corpus, test_data, bucket=bucket)
    return mm



import logging
from datetime import datetime

from emmaa.model_tests import load_model_manager_from_s3
from emmaa.db import get_db
from emmaa.util import make_date_str, find_latest_s3_file, EmailHtmlBody, \
    EMMAA_BUCKET_NAME, FORMATTED_TYPE_NAMES
from emmaa.subscription.email_util import generate_unsubscribe_link


logger = logging.getLogger(__name__)


model_manager_cache = {}


email_html = EmailHtmlBody()


class QueryManager(object):
    """Manager to run queries and interact with the database.

    Parameters
    ----------
    db : emmaa.db.EmmaaDatabaseManager
        An instance of a database manager to use.
    model_managers : list[emmaa.model_tests.ModelManager]
        Optional list of ModelManagers to use for running queries. If not
        given, the methods will load ModelManager from S3 when needed.
    """
    def __init__(self, db=None, model_managers=None):
        self.db = db
        if db is None:
            self.db = get_db('primary')
        self.model_managers = model_managers if model_managers else []

    def answer_immediate_query(
            self, user_email, user_id, query, model_names, subscribe,
            use_kappa=False, bucket=EMMAA_BUCKET_NAME):
        """This method first tries to find saved result to the query in the
        database and if not found, runs ModelManager method to answer query."""
        query_type = query.get_type()
        # Retrieve query-model hashes
        query_hashes = [
            query.get_hash_with_model(model) for model in model_names]
        # Store query in the database for future reference.
        self.db.put_queries(user_email, user_id, query, model_names, subscribe)
        # Check if the query has already been answered for any of given models
        # and retrieve the results from database.
        saved_results = self.db.get_results_from_query(query, model_names)
        if not saved_results:
            saved_results = []
        checked_models = {res[0] for res in saved_results}
        # If the query was answered for all models before, return the hashes.
        if checked_models == set(model_names):
            return {query_type: query_hashes}
        # Run queries mechanism for models for which result was not found.
        new_results = []
        new_date = datetime.now()
        for model_name in model_names:
            if model_name not in checked_models:
                results_to_store = []
                mm = self.get_model_manager(model_name)
                response_list = mm.answer_query(
                    query, use_kappa=use_kappa, bucket=bucket)
                for (mc_type, response) in response_list:
                    results_to_store.append((query, mc_type, response))
                self.db.put_results(model_name, results_to_store)
        return {query_type: query_hashes}

    def answer_registered_queries(self, model_name, find_delta=True,
                                  use_kappa=False, bucket=EMMAA_BUCKET_NAME):
        """Retrieve and asnwer registered queries

        Retrieve queries registered on database for a given model,
        answer them, calculate delta between results and put results to a
        database.

        Parameters
        ----------
        model_name : str
            The name of the model
        find_delta : bool
            If True, also generate a report for the logs on what the delta
            is for the new results compared to the previous results.
            Default: True
        use_kappa : bool
            If True, use kappa modeling when answering the dynamic query
        bucket : str
            The bucket to save the results to
        """
        model_manager = self.get_model_manager(model_name)
        queries = self.db.get_queries(model_name)
        logger.info(f'Found {len(queries)} queries for {model_name} model.')
        # Only do the following steps if there are queries for this model
        if queries:
            results = model_manager.answer_queries(
                queries, use_kappa=use_kappa, bucket=bucket)
            new_results = [(model_name, result[0], result[1], result[2], '')
                           for result in results]
            # Optionally find delta between results
            # NOTE: For now the report is presented in the logs. In future we
            # can choose some other ways to keep track of result changes.
            if find_delta:
                reports = self.make_reports_from_results(new_results, False,
                                                         'str')
                for report in reports:
                    logger.info(report)
            self.db.put_results(model_name, results)

    def get_registered_queries(self, user_email, query_type='path_property'):
        """Get formatted results to queries registered by user."""
        results = self.db.get_results(user_email, query_type=query_type)
        return format_results(results, query_type)

    def retrieve_results_from_hashes(
            self, query_hashes, query_type='path_property', latest_order=1):
        """Retrieve results from a db given a list of query-model hashes."""
        results = self.db.get_results_from_hashes(
            query_hashes, latest_order=latest_order)
        return format_results(results, query_type)

    def make_reports_from_results(
            self, new_results, stored=True, report_format='str',
            include_no_diff=True, domain='emmaa.indra.bio'):
        """Make a report given latest results and queries the results are for.

        Parameters
        ----------
        new_results : list[tuple]
            Latest results as a list of tuples where each tuple has the format
            (model_name, query, mc_type, result_json, date).
        stored : bool
            Whether the new_results are already stored in the database.
        report_format : str
            A format to write reports in. Accepted values: 'str' and 'html'.
            Default: 'str'.
        include_no_diff : bool
            If True, also report results that haven't changed. Default: True.

        Returns
        -------
        reports : list
            A list of reports on changes for each of the queries.
        """
        if report_format not in {'html', 'str'}:
            raise ValueError('Argument report_format must be either "html" '
                             'or "str"')
        processed_query_mc = []
        reports = [] if report_format == 'str' else [[], [], []]
        # If latest results are in db, retrieve the second latest
        if stored:
            order = 2
        # If latest results are not in db, retrieve the latest stored
        else:
            order = 1
        for model_name, query, mc_type, new_result_json, _ in new_results:
            if (model_name, query, mc_type) in processed_query_mc:
                continue
            try:
                old_results = self.db.get_results_from_query(
                    query, [model_name], order)
                if old_results:
                    for old_model_name, old_query, old_mc_type,\
                            old_result_json, _ in old_results:
                        if mc_type == old_mc_type and \
                                is_query_result_diff(new_result_json,
                                                     old_result_json):
                            if report_format == 'str':
                                report = self.make_str_report_one_query(
                                    model_name, query, mc_type,
                                    new_result_json, old_result_json,
                                    include_no_diff=include_no_diff)
                                if report:
                                    reports.append(report)
                            elif report_format == 'html':
                                model_type_name = FORMATTED_TYPE_NAMES[
                                    mc_type] if mc_type else mc_type
                                delta = [
                                    query.to_english(),
                                    _detailed_page_link(
                                        domain,
                                        model_name,
                                        mc_type,
                                        query.get_hash_with_model(
                                            model_name)),
                                    model_name,
                                    model_type_name
                                ]
                                # static
                                if query.get_type() == 'path_property':
                                    reports[0].append(delta)
                                # open
                                elif query.get_type() == 'open_search_query':
                                    reports[1].append(delta)
                                # dynamic
                                else:
                                    # Remove link for dynamic
                                    _ = delta.pop(1)
                                    reports[2].append(delta)
                elif report_format == 'str':
                    logger.info('No previous result was found.')
                    report = self.make_str_report_one_query(
                        model_name, query, mc_type, new_result_json,
                        None, include_no_diff=include_no_diff)
                    if report:
                        reports.append(report)
            except IndexError:
                logger.info('No result for desired date back was found.')
                if report_format == 'str':
                    report = self.make_str_report_one_query(
                        model_name, query, mc_type, new_result_json, None,
                        include_no_diff=include_no_diff)
                    if report:
                        reports.append(report)
            processed_query_mc.append((model_name, query, mc_type))
        return reports

    def get_user_query_delta(self, user_email, domain='emmaa.indra.bio'):
        """Produce a report for all query results per user in a given format

        Parameters
        ----------
        user_email : str
            The email of the user for which to get the report for
        domain : str
            The domain name for the unsubscibe link in the html
            report. Default: "emmaa.indra.bio".

        Returns
        -------
        tuple(str, html_str)
            A tuple with (str report, html report)
        """
        logger.info(f'Finding query delta for {user_email}')
        # Get results of user's query
        results = self.db.get_results(user_email, latest_order=1)

        # Make text report
        str_report = self.make_str_report_per_user(results,
                                                   include_no_diff=False)
        str_report = '\n'.join(str_report) if str_report else None

        # Make html report
        html_report = self.make_html_report_per_user(results, user_email,
                                                     domain=domain,
                                                     limit=None)
        html_report = html_report if html_report else None

        if html_report:
            logger.info(f'Found query delta for {user_email}')
        else:
            logger.info(f'No query delta to report for {user_email}')
        return str_report, html_report

    def get_report_per_query(self, model_name, query, format='str'):
        if format not in {'html', 'str'}:
            logger.error(f'Invalid format ({format}). Must be "str" '
                         f'or "html"')
            return None
        try:
            new_results = self.db.get_results_from_query(
                query, [model_name], latest_order=1)
        except IndexError:
            logger.info('No latest result was found.')
            return None
        if format == 'html':
            return self.make_reports_from_results(new_results, True, 'html')
        return self.make_reports_from_results(new_results, True, 'str')

    def make_str_report_per_user(self, results, filename=None,
                                 include_no_diff=True):
        """Produce a report for all query results per user in a text file."""
        reports = self.make_reports_from_results(
            results, True, 'str', include_no_diff=include_no_diff)
        if filename:
            with open(filename, 'w') as f:
                for report in reports:
                    f.write(report)
        else:
            return reports

    def make_html_report_per_user(self, results, email,
                                  domain='emmaa.indra.bio', limit=None):
        """Produce a report for all query results per user in an html file.

        Parameters
        ----------
        results : list[tuple]
            Results as a list of tuples where each tuple has the format
            (model_name, query, mc_type, result_json, date).
        email : str
            The email of the user to get the results for.
        domain : str
            The domain name for the unsubscibe link in the report. Default:
            "emmaa.indra.bio".
        limit : int
            The limit for how many results to show. With this set,
            on the first 'limit' results are processed into html, i.e. use
            results[:limit]. Default: all.

        Returns
        -------
        str
            A string containing an html document
        """
        results = results[:limit] if limit else results

        # Get the query deltas
        static_results_delta, open_results_delta, dynamic_results_delta = \
            self.make_reports_from_results(results, report_format='html',
                                           include_no_diff=False,
                                           domain=domain)

        # Generate unsubscribe link
        link = generate_unsubscribe_link(email=email, domain=domain)

        if static_results_delta or open_results_delta or dynamic_results_delta:
            return email_html.render(
                static_query_deltas=static_results_delta,
                open_query_deltas=open_results_delta,
                dynamic_query_deltas=dynamic_results_delta,
                unsub_link=link
            )
        else:
            return ''

    def make_str_report_one_query(self, model_name, query, mc_type,
                                  new_result_json, old_result_json=None,
                                  include_no_diff=True):
        """Return a string message containing information about a query and any
        change in the results.

        Parameters
        ----------
        model_name : str
            Name of model
        query : emmaa.query.Query
            The query object representing the query
        mc_type : str
            The model type
        new_result_json : dict
            The json containing the new results
        old_result_json : dict
            The json the new results are to be compared with
        include_no_diff : bool
            If True, also report results that haven't changed. Default: True.

        Returns
        -------
        str
            The string containing the report.
        """
        model_type_name = FORMATTED_TYPE_NAMES[mc_type] if mc_type else mc_type

        if is_query_result_diff(new_result_json, old_result_json):

            if not old_result_json:
                msg = f'\nThis is the first result to query ' \
                      f'{query.to_english()} in {model_name} with' \
                      f' {model_type_name} model checker.\nThe result is:'
                msg += _process_result_to_str(new_result_json, 'str')
            else:
                msg = f'\nA new result to query {query.to_english()}' \
                      f' in {model_name} was found with {model_type_name}' \
                      f' model checker. '
                msg += '\nPrevious result was:'
                msg += _process_result_to_str(old_result_json, 'str')
                msg += '\nNew result is:'
                msg += _process_result_to_str(new_result_json, 'str')
        elif include_no_diff:
            msg = f'\nA result to query {query.to_english()} in ' \
                  f'{model_name} from {model_type_name} model checker ' \
                  f'did not change. The result is:'
            msg += _process_result_to_str(new_result_json, 'str')
        else:
            msg = None
        return msg

    def get_model_manager(self, model_name):
        # Try get model manager from class attributes or load from s3.
        for mm in self.model_managers:
            if mm.model.name == model_name:
                return mm
        return load_model_manager_from_cache(model_name)

    def _recreate_db(self):
        self.db.drop_tables(force=True)
        self.db.create_tables()


def is_query_result_diff(new_result_json, old_result_json=None):
    """Return True if there is a delta between results."""
    # NOTE: this function is query-type specific so it may need to be
    # refactored as a method of the Query class:

    # Return True if this is the first result
    if not old_result_json:
        return True
    # Compare hashes of query results
    old_result_hashes = [k for k in old_result_json.keys()]
    new_result_hashes = [k for k in new_result_json.keys()]
    return not set(new_result_hashes) == set(old_result_hashes)


def _detailed_page_link(domain, model_name, model_type, query_hash):
    # example:
    # https://emmaa.indra.bio/query/aml/?model_type=pysb&query_hash
    # =4911955502409811&order=1
    return f'https://{domain}/query/{model_name}?model_type=' \
           f'{model_type}&query_hash={query_hash}&order=1'


def format_results(results, query_type='path_property'):
    """Format db output to a standard json structure."""
    model_types = ['pysb', 'pybel', 'signed_graph', 'unsigned_graph']
    formatted_results = {}
    for result in results:
        model = result[0]
        query = result[1]
        query_hash = query.get_hash_with_model(model)
        if query_hash not in formatted_results:
            formatted_results[query_hash] = {
                'query': query.to_english(),
                'model': model,
                'date': make_date_str(result[4])}
        mc_type = result[2]
        response_json = result[3]
        response = []
        for v in response_json.values():
            if isinstance(v, str):
                response = v
            elif isinstance(v, dict):
                response.append(v)
        if query_type in ['path_property', 'open_search_query']:
            if mc_type == '' and \
                    response == 'Query is not applicable for this model':
                for mt in model_types:
                    formatted_results[query_hash][mt] = ['n_a', response]
            elif isinstance(response, str) and \
                    response == 'Statement type not handled':
                formatted_results[query_hash][mc_type] = ['n_a', response]
            elif isinstance(response, str) and \
                    not response == 'Path found but exceeds search depth':
                formatted_results[query_hash][mc_type] = ['Fail', response]
            else:
                formatted_results[query_hash][mc_type] = ['Pass', response]
        elif query_type == 'dynamic_property':
            if response == 'Query is not applicable for this model':
                formatted_results[query_hash]['result'] = ['n_a', response]
            else:
                res = int(response[0]['sat_rate'] * 100)
                expl = (f'Satisfaction rate is {res}% after '
                        f'{response[0]["num_sim"]} simulations.')
                if res > 50:
                    formatted_results[query_hash]['result'] = ['Pass', expl]
                else:
                    formatted_results[query_hash]['result'] = ['Fail', expl]
                formatted_results[query_hash]['image'] = (
                    response[0]['fig_path'])
    if query_type in ['path_property', 'open_search_query']:
        # Loop through the results again to make sure all model types are there
        for qh in formatted_results:
            for mt in model_types:
                if mt not in formatted_results[qh]:
                    formatted_results[qh][mt] = [
                        'n_a', 'Model type not supported']
    return formatted_results


def load_model_manager_from_cache(model_name, bucket=EMMAA_BUCKET_NAME):
    model_manager = model_manager_cache.get(model_name)
    if model_manager:
        latest_on_s3 = find_latest_s3_file(
            bucket, f'results/{model_name}/model_manager_', '.pkl')
        cached_date = model_manager.date_str
        logger.info(f'Found model manager cached on {cached_date} and '
                    f'latest file on S3 is {latest_on_s3}')
        if cached_date in latest_on_s3:
            logger.info(f'Loaded model manager for {model_name} from cache.')
            return model_manager
    logger.info(f'Loading model manager for {model_name} from S3.')
    model_manager = load_model_manager_from_s3(
        model_name=model_name, bucket=bucket)
    model_manager_cache[model_name] = model_manager
    return model_manager


def _process_result_to_str(result_json, format='str'):
    msg = '\n' if format == 'str' else '<br>'
    for v in result_json.values():
        if isinstance(v, str):
            msg += v
        elif isinstance(v, dict):
            if 'path' in v.keys():
                msg += v['path']
                msg += '\n' if format == 'str' else '<br>'
            else:
                msg += f'Satisfaction rate: {v["sat_rate"]}, '
                msg += f'Number of simulations: {v["num_sim"]}, '
                msg += f'Suggested pattern: {v["kpat"]}.'
    return msg


def answer_queries_from_s3(model_name, db=None, bucket=EMMAA_BUCKET_NAME):
    """Answer registered queries with model manager on s3.

    Parameters
    ----------
    model_name : str
        Name of EmmaaModel to answer queries for.
    db : Optional[emmaa.db.manager.EmmaaDatabaseManager]
        If given over-rides the default primary database.
    """
    mm = load_model_manager_from_s3(model_name=model_name, bucket=bucket)
    qm = QueryManager(db=db, model_managers=[mm])
    qm.answer_registered_queries(model_name, find_delta=True)



import os
import logging
import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)

email_profile = 'indralabs-email'
email_bucket = 'emmaa-notifications'
notifications_sender_default = 'emmaa_notifications@indra.bio'
notifications_return_default = 'feedback@indra.bio'
indra_bio_ARN_id = os.environ.get('INDRA_BIO_ARN')


def _get_ses_client(region='us-east-1'):
    # First look for keys in environ:
    ACCESS_KEY = os.environ.get('AWS_ACCESS_KEY_ID')
    SECRET_KEY = os.environ.get('AWS_SECRET_ACCESS_KEY')
    if ACCESS_KEY and SECRET_KEY:
        logger.info('Using access keys from environment variables to get '
                    'ses client.')
        ses = boto3.session.Session(
            aws_access_key_id=ACCESS_KEY,
            aws_secret_access_key=SECRET_KEY).client(
            'ses', region_name=region)
    # If not, try to get the email_profile from the AWS credentials file
    else:
        ses = boto3.session.Session(
            profile_name=email_profile).client(
            'ses', region_name=region)
        logger.info('Using AWS credentials file to get ses client.')
    return ses


def send_email(sender, recipients, subject, body_text, body_html,
               source_arn=indra_bio_ARN_id, return_email=None,
               return_arn=None, region='us-east-1'):
    """Wrapper function for the send_email method of the boto3 SES client

    IMPORTANT: sending is limited to 14 emails per second.

    See more at:
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference +
    /services/ses.html#SES.Client.send_email
    https://docs.aws.amazon.com/ses/latest/APIReference/API_SendEmail.html
    and python example at
    https://docs.aws.amazon.com/ses/latest/DeveloperGuide/ +
    sending-authorization-delegate-sender-tasks-email.html


    Parameters
    ----------
    sender : str
        A valid email address to use in the Source field
    recipients : iterable[str] or str
        A valid email address or a list of valid email addresses. This will
        fill out the Recipients field.
    subject : str
        The email subject
    body_text : str
        The text body of the email
    body_html : str
        The html body of the email. Must be a valid html body (starting
        with <html>, ending with </html>).
    source_arn : str
        The source ARN of the sender. Should be of the format
        "arn:aws:ses:us-east-1:123456789012:identity/user@example.com" or
        "arn:aws:ses:us-east-1:123456789012:identity/example.com".
        Used only for sending authorization. It is the ARN of the identity
        that is associated with the sending authorization policy that
        permits the sender to send using the email address specified as the
        sender. Example: the owner of the domain "example.com" can send an
        email from any address using @example.com, as long as the associated
        source_arn is
        "arn:aws:ses:us-east-1:123456789012:identity/example.com"
    return_email : str
        The email to which complaints and bounces are sent. Can be the same
        as the sender.
    return_arn : str
        The return path ARN for the sender. This is the ARN associated
        with the return email. Can be the same as the source_arn if return
        email is the same as the sender.
    region : str
        AWS region to use for the SES client. Default: us-east-1

    Returns
    -------
    dict
        The API response object in the form of a dict is returned. The
        structure is:

        >>> response = {\
                'MessageId': 'EXAMPLE78603177f-7a5433e7-8edb-42ae-af10' +\
                             '-f0181f34d6ee-000000',\
                'ResponseMetadata': {\
                    '...': '...',\
                },\
            }
    """
    # Check if there is any source ARN
    if not source_arn:
        source_arn = os.environ.get('INDRA_BIO_ARN')
        if source_arn is None:
            logger.error('No SourceArn found, please set it using '
                         'the environment variable INDRA_BIO_ARN or '
                         'provided source_arn in arguments.')
            raise ValueError('source_arn must be provided or specified in '
                             'environment.')
        logger.info('Found SourceArn in os environment variable'
                    'INDRA_BIO_ARN')

    # The character encoding for the email.
    charset = "UTF-8"

    # Create a new SES client with the email profile
    ses = _get_ses_client(region)

    if return_arn is None:
        return_arn = source_arn

    if return_email is None:
        return_email = sender

    to_addresses = [rec for rec in recipients] if isinstance(
        recipients, (list, tuple, set)) else [recipients]

    # Try to send the email.
    try:
        # Provide the contents of the email.
        response = ses.send_email(
            Destination={
                'ToAddresses': to_addresses,
            },
            Message={
                'Body': {
                    'Html': {
                        'Charset': charset,
                        'Data': body_html,
                    },
                    'Text': {
                        'Charset': charset,
                        'Data': body_text,
                    },
                },
                'Subject': {
                    'Charset': charset,
                    'Data': subject,
                },
            },
            Source=sender,
            ReturnPath=return_email,
            SourceArn=source_arn,
            ReturnPathArn=return_arn
        )
    # Log error if something goes wrong.
    except ClientError as e:
        logger.error(f'Failed to send email to {", ".join(to_addresses)}')
        logger.error(e.response['Error']['Message'])
        response = e.response
    else:
        logger.info(f'Email sent to {", ".join(to_addresses)} successfully'),
    return response


def _get_max_24h_send(ses_client):
    res = ses_client.get_send_quota()
    return int(res['Max24HourSend'])


def _get_sent_last_24h(ses_client):
    res = ses_client.get_send_quota()
    return int(res['SentLast24Hours'])


def _get_quota_sent_max_ratio(ses_client):
    res = ses_client.get_send_quota()
    return res['SentLast24Hours']/res['Max24HourSend']


def close_to_quota_max(used_quota=0.95, region='us-east-1'):
    """Check if the send quota is close to be exceeded

    If the total quota for the 24h cycle is Q, the currently used quota is q
    and 'used_quota' is r, return True if q/Q > r, otherwise return False.

    Parameters
    ----------
    used_quota : float
        A float between 0 and 1.0. This number specifies the fraction of
        send quota currently used. Default: 0.95
    region : str
        A valid AWS region. The region to check the quota in.
        Default: us-east-1.

    Returns
    -------
    bool
        True if the quota is close to be exceeded with respect to the
        provided ratio 'used'.
    """
    ses = boto3.session.Session(
        profile_name=email_profile).client('ses', region_name=region)
    ratio_used = _get_quota_sent_max_ratio(ses)
    return ratio_used > used_quota


def get_send_statistics(region='us-east-1'):
    """Return the sending statistics, like bounce and complaint rates

    See
    https://boto3.amazonaws.com/v1/documentation/api/latest/
    reference/services/ses.html#SES.Client.get_send_statistics
    for more info

    Parameters
    ----------
    region : Optional[str]
        Specify AWS region

    Returns
    -------
    dict
        Response syntax:
            {
                'SendDataPoints': [
                    {
                        'Timestamp': datetime(2015, 1, 1),
                        'DeliveryAttempts': 123,
                        'Bounces': 123,
                        'Complaints': 123,
                        'Rejects': 123
                    },
                ]
            }
    """
    ses = _get_ses_client(region)
    return ses.get_send_statistics()


if __name__ == '__main__':
    logger.info('Running base case test of email')
    email_subj = input('Email subject line: ')
    msg = input('Provide a personalized message for the email body: ')
    ses_options = {
        'sender': notifications_sender_default,
        'recipients': input('email recipients (space separated): ').split(),
        'subject': email_subj,
        'body_text': f'{email_subj}\r\n'
                     'This email was sent with Amazon SES using the AWS SDK '
                     'for Python (Boto). Personal message: %s' % msg,
        'body_html': '''<html>
<head></head>
<body>
  <h1>%s</h1>
  <p>This email was sent with
    <a href='https://aws.amazon.com/ses/'>Amazon SES</a> using the
    <a href='https://aws.amazon.com/sdk-for-python/'>
      AWS SDK for Python (Boto)</a>. Personal message: %s</p>
</body>
</html>''' % (email_subj, msg),
        'source_arn': input('Provide source (sender) arn: '),
        'return_email': input('Specify ReturnPath: '),
        'return_arn': input('Specify ReturnPathArn: ')
    }
    resp = send_email(**ses_options)
    print(repr(resp))



from nose.plugins.attrib import attr

from indra.statements import Activation, Agent
from indra.explanation.model_checker import PathResult, PysbModelChecker, \
    PybelModelChecker, SignedGraphModelChecker, UnsignedGraphModelChecker
from indra.statements.statements import Statement, Agent
from emmaa.model import EmmaaModel
from emmaa.model_tests import StatementCheckingTest, ModelManager, \
    ScopeTestConnector, TestManager
from emmaa.analyze_tests_results import TestRound, StatsGenerator
from emmaa.tests.test_model import create_model


# Tell nose to not run tests in the imported modules
StatementCheckingTest.__test__ = False
TestRound.__test__ = False
ScopeTestConnector.__test__ = False
TestManager.__test__ = False


def test_model_manager_structure():
    model = create_model()
    mm = ModelManager(model)
    assert isinstance(mm, ModelManager)
    assert isinstance(mm.model, EmmaaModel)
    assert mm.model.name == 'test'
    assert len(mm.mc_types) == 4, len(mm.mc_types)
    assert len(mm.mc_types['pysb']) == 3
    assert isinstance(mm.mc_types['pysb']['model_checker'], PysbModelChecker)
    assert isinstance(mm.mc_types['pybel']['model_checker'], PybelModelChecker)
    assert isinstance(
        mm.mc_types['signed_graph']['model_checker'], SignedGraphModelChecker)
    assert isinstance(
        mm.mc_types['unsigned_graph']['model_checker'],
        UnsignedGraphModelChecker)
    assert isinstance(mm.entities[0], Agent)
    assert isinstance(mm.date_str, str)


def test_run_tests():
    model = create_model()
    tests = [StatementCheckingTest(
             Activation(Agent('BRAF', db_refs={'HGNC': '1097'}),
                        Agent('MAPK1', db_refs={'UP': 'P28482'})))]
    mm = ModelManager(model)
    tm = TestManager([mm], tests)
    tm.make_tests(ScopeTestConnector())
    tm.run_tests()
    assert len(mm.applicable_tests) == 1
    assert isinstance(mm.applicable_tests[0], StatementCheckingTest)
    assert len(mm.mc_types['pysb']['test_results']) == 1
    assert len(mm.mc_types['pybel']['test_results']) == 1
    assert len(mm.mc_types['signed_graph']['test_results']) == 1
    assert len(mm.mc_types['unsigned_graph']['test_results']) == 1
    assert isinstance(mm.mc_types['pysb']['test_results'][0], PathResult)



"""The AWS Lambda emmaa-after-update definition.

This file contains the function that will be run when Lambda is triggered. It
must be placed on s3, which can either be done manually (not recommended) or
by running:

$ python update_lambda.py after_update.py emmaa-after-update

in this directory.
"""

import boto3
import json
from datetime import datetime


batch = boto3.client('batch')
JOB_DEF = 'emmaa_jobdef'
QUEUE = 'emmaa-models-update-test'
PROJECT = 'aske'
BRANCH = 'origin/master'
now_str = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
date = datetime.utcnow().strftime('%Y-%m-%d')


def submit_batch_job(script_command, purpose, job_name, wait_for=None):
    print(f'Submitting job {job_name}')
    core_command = 'bash scripts/git_and_run.sh'
    if BRANCH is not None:
        core_command += f' --branch {BRANCH}'
    core_command += script_command
    print(core_command)
    cont_overrides = {
        'command': ['python', '-m', 'indra.util.aws', 'run_in_batch',
                    '--project', PROJECT, '--purpose', purpose,
                    core_command]
        }
    kwargs = {}
    if wait_for:
        kwargs['dependsOn'] = [{'jobId': job_id, 'type': 'SEQUENTIAL'}
                               for job_id in wait_for]
    ret = batch.submit_job(
        jobName=job_name,
        jobQueue=QUEUE, jobDefinition=JOB_DEF,
        containerOverrides=cont_overrides, **kwargs)
    job_id = ret['jobId']
    print(f"Result from job submission: {job_id}")
    return job_id


def lambda_handler(event, context):
    """Submit model tests, model and test stats, and query batch jobs.

    This function is designed to be placed on AWS Lambda, taking the event and
    context arguments that are passed. Note that this function must always have
    the same parameters, even if any or all of them are unused, because we do
    not have control over what Lambda sends as parameters. Event parameter is
    used here to pass which model manager was updated.

    Lambda is configured to run this script when ModelManager object is
    updated.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event. In
        this case, we are expecting 'Records', each of which contains a record
        of a file that was added (or changed) on s3.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A dict containing 'statusCode', with a valid HTTP status code, and any
        other data to be returned to Lambda.
    """
    s3 = boto3.client('s3')
    records = event['Records']
    for rec in records:
        try:
            model_key = rec['s3']['object']['key']
        except KeyError:
            pass
        model_name = model_key.split('/')[1]

        # Store all stats jobs IDs
        stats_job_ids = []

        # Submit model stats job
        model_stats_command = (' python scripts/run_model_stats_from_s3.py'
                               f' --model {model_name}  --stats_mode model')
        model_stats_id = submit_batch_job(
            model_stats_command, 'update-emmaa-model-stats',
            f'{model_name}_model_stats_{now_str}')
        stats_job_ids.append(model_stats_id)

        # Find all test corpora for daily runi
        config_key = f'models/{model_name}/config.json'
        obj = s3.get_object(Bucket='emmaa', Key=config_key)
        config = json.loads(obj['Body'].read().decode('utf8'))
        tests = config['test'].get('test_corpus', 'large_corpus_tests')
        if isinstance(tests, str):
            tests = [tests]

        # For each test run the test and test stats
        for test_corpus in tests:
            test_command = (' python scripts/run_model_tests_from_s3.py'
                            f' --model {model_name} --tests {test_corpus}')
            test_id = submit_batch_job(
                test_command, 'update-emmaa-results',
                f'{model_name}_{test_corpus}_tests_{now_str}')
            test_stats_command = (' python scripts/run_model_stats_from_s3.py'
                                  f' --model {model_name} --stats_mode tests'
                                  f' --tests {test_corpus}')
            test_stats_id = submit_batch_job(
                test_stats_command, 'update-emmaa-test-stats',
                f'{model_name}_{test_corpus}_stats_{now_str}', [test_id])
            stats_job_ids.append(test_stats_id)

        # Submit twitter job
        if config.get('twitter'):
            twitter_command = (
                f' python scripts/tweet_deltas.py --model {model_name} '
                f'--test_corpora {" ".join(tc for tc in tests)} --date {date}')
            submit_batch_job(twitter_command, 'update-twitter-status',
                             f'{model_name}_twitter_{now_str}', stats_job_ids)
        # Run queries
        query_command = (' python scripts/answer_queries_from_s3.py'
                         f' --model {model_name}')
        submit_batch_job(query_command, 'update-emmaa-queries',
                         f'{model_name}_queries_{now_str}')

        # Make tests if configured
        if config.get('make_tests', False):
            test_update_command = (' python scripts/model_to_tests.py'
                                   f' --model {model_name}')
            submit_batch_job(test_update_command, 'update-emmaa-tests',
                             f'{model_name}_test_update_{now_str}')

    return 'All jobs sumbitted'



from os.path import abspath, dirname, join
from datetime import datetime
from nose.plugins.attrib import attr
from emmaa.answer_queries import QueryManager, format_results, \
    is_query_result_diff
from emmaa.queries import Query, DynamicProperty, get_agent_from_trips
from emmaa.model_tests import ModelManager
from emmaa.tests.test_db import _get_test_db
from emmaa.tests.test_model import create_model


test_query = {'type': 'path_property', 'path': {'type': 'Activation',
              'subj': {'type': 'Agent', 'name': 'BRAF',
                       'db_refs': {'HGNC': '1097'}},
              'obj': {'type': 'Agent', 'name': 'MAPK1',
                      'db_refs': {'HGNC': '6871'}},
                      'obj_activity': 'activity'}}
simple_query = 'BRAF activates MAPK1.'
query_object = Query._from_json(test_query)
dyn_ag = get_agent_from_trips('active MAP2K1')
dyn_query = DynamicProperty(dyn_ag, 'eventual_value', 'high')
open_qj = {'type': 'open_search_query',
            'entity': {'type': 'Agent', 'name': 'BRAF',
                       'db_refs': {'HGNC': '1097'}},
            'entity_role': 'subject', 'stmt_type': 'Activation'}
open_query = Query._from_json(open_qj)
test_response = {
    '3801854542': {
        'path': 'BRAF → MAP2K1 → MAPK1',
        'edge_list': [
            {'edge': 'BRAF → MAP2K1',
             'stmts': [
                 ['/evidence?stmt_hash=-23078353002754841&source='
                  'model_statement&model=test&date=2020-01-01',
                  'BRAF activates MAP2K1.', '']]},
            {'edge': 'MAP2K1 → MAPK1',
             'stmts': [
                 ['/evidence?stmt_hash=-34603994586320440&source='
                  'model_statement&model=test&date=2020-01-01',
                  'Active MAP2K1 activates MAPK1.', '']]}]}}
query_not_appl = {'2413475507': 'Query is not applicable for this model'}
fail_response = {
    '521653329': 'No path found that satisfies the test statement'}
# Create a new EmmaaModel and ModelManager for tests instead of depending
# on S3 version
test_model = create_model()
test_mm = ModelManager(test_model)
test_mm.date_str = '2020-01-01-00-00-00'
test_email = 'tester@test.com'


def test_format_results():
    date = datetime.now()
    results = [('test', query_object, 'pysb', test_response, date),
               ('test', query_object, 'signed_graph', fail_response, date),
               ('test', query_object, 'unsigned_graph', test_response, date)]
    formatted_results = format_results(results)
    assert len(formatted_results) == 1
    qh = query_object.get_hash_with_model('test')
    assert qh in formatted_results
    assert formatted_results[qh]['model'] == 'test'
    assert formatted_results[qh]['query'] == simple_query
    assert isinstance(formatted_results[qh]['date'], str)
    assert formatted_results[qh]['pysb'] == [
        'Pass', [test_response['3801854542']]]
    assert formatted_results[qh]['pybel'] == [
        'n_a', 'Model type not supported']
    assert formatted_results[qh]['signed_graph'] == [
        'Fail', fail_response['521653329']]
    assert formatted_results[qh]['unsigned_graph'] == [
        'Pass', [test_response['3801854542']]]


@attr('nonpublic')
def test_answer_immediate_query():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    query_hashes = qm.answer_immediate_query(
        test_email, 1, query_object, ['test'], subscribe=False)[
            'path_property']
    assert query_hashes == [35683418474694258], query_hashes
    results = qm.retrieve_results_from_hashes(query_hashes)
    assert len(results) == 1
    assert query_hashes[0] in results
    result_values = results[query_hashes[0]]
    assert result_values['model'] == 'test'
    assert result_values['query'] == simple_query
    assert isinstance(result_values['date'], str)
    assert result_values['pysb'] == ['Pass', [test_response['3801854542']]], \
        result_values['pysb']
    assert result_values['pybel'] == ['Pass', [test_response['3801854542']]]
    assert result_values['signed_graph'][0] == 'Pass'
    assert result_values['unsigned_graph'][0] == 'Pass'


@attr('nonpublic')
def test_immediate_dynamic():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    query_hashes = qm.answer_immediate_query(
        test_email, 1, dyn_query, ['test'], subscribe=False)[
            'dynamic_property']
    assert query_hashes == [-27775603206605897], query_hashes
    results = qm.retrieve_results_from_hashes(query_hashes, 'dynamic_property')
    assert len(results) == 1, results
    assert query_hashes[0] in results
    result_values = results[query_hashes[0]]
    assert result_values['model'] == 'test'
    assert result_values['query'] == 'Active MAP2K1 is eventually high.'
    assert isinstance(result_values['date'], str)
    assert result_values['result'] == [
        'Pass', 'Satisfaction rate is 100% after 2 simulations.']
    assert isinstance(result_values['image'], str)


def test_immediate_open():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    query_hashes = qm.answer_immediate_query(
        test_email, 1, open_query, ['test'], subscribe=False)[
            'open_search_query']
    assert query_hashes == [-13552944417558866], query_hashes
    results = qm.retrieve_results_from_hashes(query_hashes)
    assert len(results) == 1
    assert query_hashes[0] in results
    result_values = results[query_hashes[0]]
    assert result_values['model'] == 'test'
    assert result_values['query'] == 'What does BRAF activate?'
    assert isinstance(result_values['date'], str)
    for mc_type in ['pysb', 'pybel', 'signed_graph', 'unsigned_graph']:
        assert result_values[mc_type][0] == 'Pass'
        assert isinstance(result_values[mc_type][1], list)
        assert test_response['3801854542']['path'] in [
            res['path'] for res in result_values[mc_type][1]]


@attr('nonpublic')
def test_answer_get_registered_queries():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    # Put all types of queries in db, answer together
    qm.db.put_queries(test_email, 1, query_object, ['test'],
                      subscribe=True)
    qm.db.put_queries(test_email, 1, dyn_query, ['test'],
                      subscribe=True)
    qm.db.put_queries(test_email, 1, open_query, ['test'],
                      subscribe=True)
    qm.answer_registered_queries('test')
    # Retrieve results for path query
    results = qm.get_registered_queries(test_email, 'path_property')
    qh = query_object.get_hash_with_model('test')
    assert qh in results
    assert len(results) == 1
    assert results[qh]['model'] == 'test'
    assert results[qh]['query'] == simple_query
    assert isinstance(results[qh]['date'], str)
    assert results[qh]['pysb'] == ['Pass', [test_response['3801854542']]], \
        (results[qh]['pysb'], test_response['3801854542'])
    assert results[qh]['pybel'] == ['Pass', [test_response['3801854542']]]
    assert results[qh]['signed_graph'][0] == 'Pass'
    assert results[qh]['unsigned_graph'][0] == 'Pass'
    # Retrieve results for dynamic query
    results = qm.get_registered_queries(test_email, 'dynamic_property')
    qh = dyn_query.get_hash_with_model('test')
    assert qh in results
    assert results[qh]['model'] == 'test'
    assert results[qh]['query'] == 'Active MAP2K1 is eventually high.'
    assert isinstance(results[qh]['date'], str)
    assert results[qh]['result'] == [
        'Pass', 'Satisfaction rate is 100% after 2 simulations.']
    assert isinstance(results[qh]['image'], str)
    # Retrieve results for dynamic query
    results = qm.get_registered_queries(test_email, 'open_search_query')
    qh = open_query.get_hash_with_model('test')
    assert qh in results
    assert results[qh]['model'] == 'test'
    assert results[qh]['query'] == 'What does BRAF activate?'
    assert isinstance(results[qh]['date'], str)
    for mc_type in ['pysb', 'pybel', 'signed_graph', 'unsigned_graph']:
        assert results[qh][mc_type][0] == 'Pass'
        assert isinstance(results[qh][mc_type][1], list)
        assert test_response['3801854542']['path'] in [
            res['path'] for res in results[qh][mc_type][1]]


def test_is_diff():
    assert not is_query_result_diff(query_not_appl, query_not_appl)
    assert is_query_result_diff(test_response, query_not_appl)


@attr('nonpublic')
def test_report_one_query():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    # Using results from db
    qm.db.put_queries(test_email, 1, query_object, ['test'],
                      subscribe=True)
    qm.db.put_results('test', [(query_object, 'pysb', test_response),
                               (query_object, 'pysb', query_not_appl)])
    str_msg = qm.get_report_per_query('test', query_object)[0]
    assert str_msg
    assert 'A new result to query' in str_msg
    assert 'Query is not applicable for this model' in str_msg
    assert 'BRAF → MAP2K1 → MAPK1' in str_msg
    # String report given two responses explicitly
    str_msg = qm.make_str_report_one_query(
        'test', query_object, 'pysb', test_response, query_not_appl)
    assert str_msg
    assert 'A new result to query' in str_msg, str_msg
    assert 'Query is not applicable for this model' in str_msg
    assert 'BRAF → MAP2K1 → MAPK1' in str_msg
    assert simple_query in str_msg


@attr('nonpublic')
def test_report_files():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    qm.db.put_queries(test_email, 1, query_object, ['test'],
                      subscribe=True)
    qm.db.put_results('test', [(query_object, 'pysb', query_not_appl)])
    results = qm.db.get_results(test_email, latest_order=1)
    qm.make_str_report_per_user(results,
                                filename='test_query_delta.txt')
    report_file = join(dirname(abspath(__file__)), 'test_query_delta.txt')
    with open(report_file, 'r') as f:
        msg = f.read()
    assert msg
    assert 'This is the first result to query' in msg, msg
    assert 'Query is not applicable for this model' in msg
    qm.db.put_results('test', [(query_object, 'pysb', test_response)])
    results = qm.db.get_results(test_email, latest_order=1)
    qm.make_str_report_per_user(results,
                                filename='new_test_query_delta.txt')
    new_report_file = join(dirname(abspath(__file__)),
                           'new_test_query_delta.txt')
    with open(new_report_file, 'r') as f:
        msg = f.read()
    assert msg
    assert 'A new result to query' in msg
    assert 'BRAF → MAP2K1 → MAPK1' in msg


@attr('nonpublic')
def test_user_query_delta():
    db = _get_test_db()
    qm = QueryManager(db=db, model_managers=[test_mm])
    # Using results from db
    qm.db.put_queries(test_email, 1, query_object, ['test'],
                      subscribe=True)
    qm.db.put_results('test', [(query_object, 'pysb', test_response)])
    qm.db.put_results('test', [(query_object, 'pysb', query_not_appl)])
    str_rep, html_rep = qm.get_user_query_delta(user_email=test_email)
    assert str_rep, print(str_rep)
    assert html_rep, print(html_rep)
    assert '</html>' in html_rep
    assert '</body>' in html_rep
    assert 'pysb' in html_rep
    assert 'unsubscribe' in html_rep



"""The AWS Lambda emmaa-test-pipeline definition.

This file contains the function that will be run when Lambda is triggered. It
must be placed on s3, which can either be done manually (not recommended) or
by running:

$ python update_lambda.py test_pipeline.py emmaa-test-pipeline

in this directory.
"""

import boto3
import json


def lambda_handler(event, context):
    """Invoke individual test corpus functions.

    This function is designed to be placed on lambda, taking the event and
    context arguments that are passed. Event parameter is used here to pass
    name of the model.

    This Lambda function is configured to be invoked by emmaa-after-update
    Lambda function.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event. In
        this case the dictionary contains 'model' key.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A response returned by the latest call to emmaa-model-test function.
    """
    s3 = boto3.client('s3')
    lam = boto3.client('lambda')
    model_name = event['model']
    config_key = f'models/{model_name}/config.json'
    obj = s3.get_object(Bucket='emmaa', Key=config_key)
    config = json.loads(obj['Body'].read().decode('utf8'))
    tests = config['test'].get('test_corpus', 'large_corpus_tests')
    if isinstance(tests, str):
        resp = lam.invoke(FunctionName='emmaa-model-test',
                          InvocationType='RequestResponse',
                          Payload=json.dumps({"model": model_name,
                                              "tests": tests}))
    elif isinstance(tests, list):
        for test in tests:
            resp = lam.invoke(FunctionName='emmaa-model-test',
                              InvocationType='RequestResponse',
                              Payload=json.dumps({"model": model_name,
                                                  "tests": test}))
    print(resp['Payload'].read())
    return {'statusCode': 200, 'result': 'SUCCESS'}



"""The AWS Lambda email-notification definition

This file contains the function that starts the email notification batch job.
It must be placed on s3, which can either be done manually (not recommended)
or by running:

$ python update_lambda.py email_notifications.py emmaa-email-notifications

in this directory.
"""

import boto3
from datetime import datetime

JOB_DEF = 'emmaa-email-notifications'
QUEUE = 'emmaa-models-update-test'
PROJECT = 'aske'
PURPOSE = 'emmaa-email-notifications'
BRANCH = 'origin/master'


def lambda_handler(event, context):
    """Invoke batch job that checks for query deltas per subscribed user

    This function is designed to be placed on AWS Lambda, taking the event and
    context arguments that are passed. Note that this function must always have
    the same parameters, even if any or all of them are unused, because we do
    not have control over what Lambda sends as parameters.

    This Lambda function is configured to be invoked daily.

    See the top of the page for the Lambda update procedure.

    Parameters
    ----------
    event : dict
        A dictionary containing metadata regarding the triggering event.
    context : object
        This is an object containing potentially useful context provided by
        Lambda. See the documentation cited above for details.

    Returns
    -------
    ret : dict
        A dict containing 'statusCode', with a valid HTTP status code,
        'result', and 'job_id' to be returned to Lambda.
    """
    batch = boto3.client('batch')
    core_command = 'bash scripts/git_and_run.sh'
    if BRANCH is not None:
        core_command += f' --branch {BRANCH} '
    core_command += 'python scripts/run_email_notifications.py'
    print(core_command)
    cont_overrides = {
        'command': ['python', '-m', 'indra.util.aws', 'run_in_batch',
                    '--project', PROJECT, '--purpose', PURPOSE,
                    core_command]
        }
    now_str = datetime.utcnow().strftime('%Y%m%d_%H%M%S')
    ret = batch.submit_job(jobName=f'email_notifications_{now_str}',
                           jobQueue=QUEUE, jobDefinition=JOB_DEF,
                           containerOverrides=cont_overrides)
    job_id = ret['jobId']

    return {'statusCode': 200, 'result': 'SUCCESS', 'job_id': job_id}
